"""
pyart.util.xsect
================

Function for extracting cross sections from radar volumes.

.. autosummary::
    :toctree: generated/

    cross_section_ppi
    cross_section_rhi
    colocated_gates
    colocated_gates2
    intersection
    find_intersection_volume
    find_intersection_limits
    find_equal_vol_region
    get_ground_distance
    get_range
    get_vol_diameter
    _construct_xsect_radar
    _copy_dic

"""

from copy import copy
from warnings import warn

import numpy as np
from scipy.spatial import cKDTree


from ..core import Radar, geographic_to_cartesian_aeqd
from ..config import get_metadata, get_field_name


def cross_section_ppi(radar, target_azimuths, az_tol=None):
    """
    Extract cross sections from a PPI volume along one or more azimuth angles.

    Parameters
    ----------
    radar : Radar
        Radar volume containing PPI sweeps from which azimuthal
        cross sections will be extracted.
    target_azimuth : list
        Azimuthal angles in degrees where cross sections will be taken.
    az_tol : float, optional
        Azimuth angle tolerance in degrees. If none the nearest angle is used.
        If valid only angles within the tolerance distance are considered.

    Returns
    -------
    radar_rhi : Radar
        Radar volume containing RHI sweeps which contain azimuthal
        cross sections from the original PPI volume.

    """
    # determine which rays from the ppi radar make up the pseudo RHI
    prhi_rays = []
    valid_azimuths = []
    rhi_sweep_start_ray_index = []
    rhi_sweep_end_ray_index = []
    ray_index = -1
    for target_azimuth in target_azimuths:
        found_first = False
        for sweep_slice in radar.iter_slice():
            sweep_azimuths = radar.azimuth['data'][sweep_slice]
            d_az = np.abs(sweep_azimuths - target_azimuth)
            if az_tol is None:
                ray_number = np.argmin(d_az)
                prhi_rays.append(ray_number + sweep_slice.start)
            else:
                d_az_min = np.min(d_az)
                if d_az_min > az_tol:
                    warn('WARNING: No azimuth found whithin tolerance ' +
                         'for angle '+str(target_azimuth) +
                         '. Minimum distance to radar azimuth ' +
                         str(d_az_min)+' larger than tolerance ' +
                         str(az_tol))
                    continue
                ray_number = np.argmin(d_az)
                prhi_rays.append(ray_number + sweep_slice.start)

            ray_index += 1
            if not found_first:
                valid_azimuths.append(target_azimuth)
                rhi_sweep_start_ray_index.append(ray_index)
                found_first = True
        if found_first:
            rhi_sweep_end_ray_index.append(ray_index)

    rhi_nsweeps = len(valid_azimuths)
    if rhi_nsweeps == 0:
        raise ValueError('No azimuth found within tolerance')

    radar_rhi = _construct_xsect_radar(
        radar, 'rhi', prhi_rays, rhi_sweep_start_ray_index,
        rhi_sweep_end_ray_index, valid_azimuths)

    return radar_rhi


def cross_section_rhi(radar, target_elevations, el_tol=None):
    """
    Extract cross sections from an RHI volume along one or more elevation
    angles.

    Parameters
    ----------
    radar : Radar
        Radar volume containing RHI sweeps from which azimuthal
        cross sections will be extracted.
    target_elevations : list
        Elevation angles in degrees where cross sections will be taken.
    el_tol : float, optional
        Elevation angle tolerance in degrees. If none the nearest angle is
        used. If valid only angles within the tolerance distance are
        considered.

    Returns
    -------
    radar_ppi : Radar
        Radar volume containing PPI sweeps which contain azimuthal
        cross sections from the original RHI volume.

    """

    # determine which rays from the rhi radar make up the pseudo PPI
    pppi_rays = []
    valid_elevations = []
    ppi_sweep_start_ray_index = []
    ppi_sweep_end_ray_index = []
    ray_index = -1
    for target_elevation in target_elevations:
        found_first = False
        for sweep_slice in radar.iter_slice():
            sweep_elevations = radar.elevation['data'][sweep_slice]
            d_el = np.abs(sweep_elevations - target_elevation)
            if el_tol is None:
                ray_number = np.argmin(d_el)
                pppi_rays.append(ray_number + sweep_slice.start)
            else:
                d_el_min = np.min(d_el)
                if d_el_min > el_tol:
                    warn('WARNING: No elevation found whithin tolerance ' +
                         'for angle '+str(target_elevation) +
                         '. Minimum distance to radar elevation ' +
                         str(d_el_min) + ' larger than tolerance ' +
                         str(el_tol))
                    continue
                ray_number = np.argmin(d_el)
                pppi_rays.append(ray_number + sweep_slice.start)

            ray_index += 1
            if not found_first:
                valid_elevations.append(target_elevation)
                ppi_sweep_start_ray_index.append(ray_index)
                found_first = True
        if found_first:
            ppi_sweep_end_ray_index.append(ray_index)

    ppi_nsweeps = len(valid_elevations)
    if ppi_nsweeps == 0:
        raise ValueError('No elevation found within tolerance')

    radar_ppi = _construct_xsect_radar(
        radar, 'ppi', pppi_rays, ppi_sweep_start_ray_index,
        ppi_sweep_end_ray_index, valid_elevations)

    return radar_ppi


def colocated_gates(radar1, radar2, h_tol=0., latlon_tol=0.,
                    coloc_gates_field=None):
    """
    Flags radar gates of radar1 colocated with radar2

    Parameters
    ----------
    radar1 : Radar
        radar object that is going to be flagged
    radar2 : Radar
        radar object
    h_tol : float
        tolerance in altitude [m]
    latlon_tol : float
        tolerance in latitude/longitude [deg]
    coloc_gates_field : string
        Name of the field to retrieve the data

    Returns
    -------
    coloc_dict : dict
        a dictionary containing the colocated positions of radar 1
        (ele, azi, rng) and radar 2
    coloc_rad1 :
        field with the colocated gates of radar1 flagged, i.e:
        1: not colocated gates 2: colocated (0 is reserved)

    """
    # parse the field parameters
    if coloc_gates_field is None:
        coloc_gates_field = get_field_name('colocated_gates')

    coloc_dict = {
        'rad1_ele': [],
        'rad1_azi': [],
        'rad1_rng': [],
        'rad1_ray_ind': [],
        'rad1_rng_ind': [],
        'rad2_ele': [],
        'rad2_azi': [],
        'rad2_rng': [],
        'rad2_ray_ind': [],
        'rad2_rng_ind': []}

    coloc_rad1 = radar1.fields[coloc_gates_field]
    coloc_rad2 = radar2.fields[coloc_gates_field]

    ind_ray_rad1, ind_rng_rad1 = np.where(coloc_rad1['data'] == 2)
    ngates = len(ind_ray_rad1)
    # debug output:
    # print('looking whether '+str(ngates) +
    #       ' gates of radar1 are colocated with radar2. ' +
    #       'This may take a while...')

    # Make region preselection for radar 2
    i_ray_psel, i_rng_psel = np.where(coloc_rad2['data'] == 2)

    # compute Cartesian position of radar1 respect to radar 2
    x0, y0 = geographic_to_cartesian_aeqd(
        radar1.longitude['data'], radar1.latitude['data'],
        radar2.longitude['data'][0], radar2.latitude['data'][0], R=6370997.)
    z0 = radar1.altitude['data'][0]-radar2.altitude['data'][0]

    for i in range(ngates):
        rad1_alt = radar1.gate_altitude['data'][
            ind_ray_rad1[i], ind_rng_rad1[i]]
        rad1_lat = radar1.gate_latitude['data'][
            ind_ray_rad1[i], ind_rng_rad1[i]]
        rad1_lon = radar1.gate_longitude['data'][
            ind_ray_rad1[i], ind_rng_rad1[i]]

        inds = np.where(
            np.logical_and(
                np.logical_and(
                    radar2.gate_altitude['data'][i_ray_psel, i_rng_psel] <
                    rad1_alt+h_tol,
                    radar2.gate_altitude['data'][i_ray_psel, i_rng_psel] >
                    rad1_alt-h_tol),
                np.logical_and(
                    np.logical_and(
                        radar2.gate_latitude['data'][i_ray_psel, i_rng_psel] <
                        rad1_lat+latlon_tol,
                        radar2.gate_latitude['data'][i_ray_psel, i_rng_psel] >
                        rad1_lat-latlon_tol),
                    np.logical_and(
                        radar2.gate_longitude['data'][i_ray_psel, i_rng_psel] <
                        rad1_lon+latlon_tol,
                        radar2.gate_longitude['data'][i_ray_psel, i_rng_psel] >
                        rad1_lon-latlon_tol)
                    )
                ))

        if inds[0].size == 0:
            # not colocated: set co-located flag to 1
            coloc_rad1['data'][ind_ray_rad1[i], ind_rng_rad1[i]] = 1
            continue

        ind_ray_rad2 = i_ray_psel[inds]
        ind_rng_rad2 = i_rng_psel[inds]

        if len(ind_ray_rad2) == 1:
            ind_ray_rad2 = ind_ray_rad2[0]
            ind_rng_rad2 = ind_rng_rad2[0]
        else:
            # compute minimum distance
            # position of radar 1 gate respect to radar 2
            rad1_x = radar1.gate_x['data'][ind_ray_rad1[i], ind_rng_rad1[i]]+x0
            rad1_y = radar1.gate_y['data'][ind_ray_rad1[i], ind_rng_rad1[i]]+y0
            rad1_z = radar1.gate_z['data'][ind_ray_rad1[i], ind_rng_rad1[i]]+z0

            rad2_x = radar2.gate_x['data'][ind_ray_rad2, ind_rng_rad2]
            rad2_y = radar2.gate_y['data'][ind_ray_rad2, ind_rng_rad2]
            rad2_z = radar2.gate_z['data'][ind_ray_rad2, ind_rng_rad2]

            dist = np.sqrt(
                (rad2_x-rad1_x)**2.+(rad2_y-rad1_y)**2.+(rad2_z-rad1_z)**2.)
            ind_min = np.argmin(dist)

            ind_ray_rad2 = ind_ray_rad2[ind_min]
            ind_rng_rad2 = ind_rng_rad2[ind_min]

        # colocated and valid gate
        coloc_dict['rad1_ele'].append(
            radar1.elevation['data'][ind_ray_rad1[i]])
        coloc_dict['rad1_azi'].append(
            radar1.azimuth['data'][ind_ray_rad1[i]])
        coloc_dict['rad1_rng'].append(
            radar1.range['data'][ind_rng_rad1[i]])
        coloc_dict['rad1_ray_ind'].append(
            ind_ray_rad1[i])
        coloc_dict['rad1_rng_ind'].append(
            ind_rng_rad1[i])
        coloc_dict['rad2_ele'].append(
            radar2.elevation['data'][ind_ray_rad2])
        coloc_dict['rad2_azi'].append(
            radar2.azimuth['data'][ind_ray_rad2])
        coloc_dict['rad2_rng'].append(
            radar2.range['data'][ind_rng_rad2])
        coloc_dict['rad2_ray_ind'].append(
            ind_ray_rad2)
        coloc_dict['rad2_rng_ind'].append(
            ind_rng_rad2)

    #     # debug output:
    #     print(
    #         radar1.elevation['data'][ind_ray_rad1[i]],
    #         radar1.azimuth['data'][ind_ray_rad1[i]],
    #         radar1.range['data'][ind_rng_rad1[i]],
    #         radar2.elevation['data'][ind_ray_rad2],
    #         radar2.azimuth['data'][ind_ray_rad2],
    #         radar2.range['data'][ind_rng_rad2])
    #     print(
    #         radar1.gate_longitude['data'][ind_ray_rad1[i], ind_rng_rad1[i]],
    #         radar1.gate_latitude['data'][ind_ray_rad1[i], ind_rng_rad1[i]],
    #         radar1.gate_altitude['data'][ind_ray_rad1[i], ind_rng_rad1[i]],
    #         radar2.gate_longitude['data'][ind_ray_rad2, ind_rng_rad2],
    #         radar2.gate_latitude['data'][ind_ray_rad2, ind_rng_rad2],
    #         radar2.gate_altitude['data'][ind_ray_rad2, ind_rng_rad2])
    #
    # ind_ray_rad1, ind_rng_rad1 = np.where(coloc_rad1['data'])
    # ngates = len(ind_ray_rad1)
    # print(str(ngates)+' gates of radar1 are colocated with radar2.')

    return coloc_dict, coloc_rad1


def colocated_gates2(radar1, radar2, distance_upper_bound=1000.,
                     coloc_gates_field=None):
    """
    Flags radar gates of radar1 co-located with radar2. Uses nearest neighbour
    calculation with cKDTree

    Parameters
    ----------
    radar1 : Radar
        radar object that is going to be flagged
    radar2 : Radar
        radar object
    distance_upper_bound : float
        upper bound of the distance between neighbours (m)
    coloc_gates_field : string
        Name of the field to retrieve the data

    Returns
    -------
    coloc_dict : dict
        a dictionary containing the colocated positions of radar 1
        (ele, azi, rng) and radar 2
    coloc_rad1 :
        field with the colocated gates of radar1 flagged, i.e:
        1: not colocated gates 2: colocated (0 is reserved)

    """
    # parse the field parameters
    if coloc_gates_field is None:
        coloc_gates_field = get_field_name('colocated_gates')

    coloc_dict = {
        'rad1_ele': [],
        'rad1_azi': [],
        'rad1_rng': [],
        'rad1_ray_ind': [],
        'rad1_rng_ind': [],
        'rad2_ele': [],
        'rad2_azi': [],
        'rad2_rng': [],
        'rad2_ray_ind': [],
        'rad2_rng_ind': []}

    coloc_rad1 = radar1.fields[coloc_gates_field]
    coloc_rad2 = radar2.fields[coloc_gates_field]

    ind_ray_rad1, ind_rng_rad1 = np.where(coloc_rad1['data'] == 2)

    # Make region pre-selection for radar 2
    i_ray_psel, i_rng_psel = np.where(coloc_rad2['data'] == 2)

    # compute Cartesian position of radar1 respect to radar 2
    x0, y0 = geographic_to_cartesian_aeqd(
        radar1.longitude['data'], radar1.latitude['data'],
        radar2.longitude['data'][0], radar2.latitude['data'][0], R=6370997.)
    z0 = radar1.altitude['data'][0]-radar2.altitude['data'][0]

    # Position of radar 1 gates respect to radar 2
    rad1_x = radar1.gate_x['data'][ind_ray_rad1, ind_rng_rad1]+x0
    rad1_y = radar1.gate_y['data'][ind_ray_rad1, ind_rng_rad1]+y0
    rad1_z = radar1.gate_z['data'][ind_ray_rad1, ind_rng_rad1]+z0

    rad2_x = radar2.gate_x['data'][i_ray_psel, i_rng_psel]
    rad2_y = radar2.gate_y['data'][i_ray_psel, i_rng_psel]
    rad2_z = radar2.gate_z['data'][i_ray_psel, i_rng_psel]

    tree = cKDTree(
        np.transpose((rad2_x, rad2_y, rad2_z)),
        compact_nodes=False, balanced_tree=False)
    dist, ind_vec = tree.query(np.transpose((rad1_x, rad1_y, rad1_z)), k=1)

    ind_not_col = np.where(dist > distance_upper_bound)
    ind_col = np.where(dist <= distance_upper_bound)
    ind_vec = ind_vec[ind_col]

    # colocated and valid gate
    coloc_dict['rad1_ele'] = radar1.elevation['data'][ind_ray_rad1[ind_col]]
    coloc_dict['rad1_azi'] = radar1.azimuth['data'][ind_ray_rad1[ind_col]]
    coloc_dict['rad1_rng'] = radar1.range['data'][ind_rng_rad1[ind_col]]
    coloc_dict['rad1_ray_ind'] = ind_ray_rad1[ind_col]
    coloc_dict['rad1_rng_ind'] = ind_rng_rad1[ind_col]
    coloc_dict['rad2_ele'] = radar2.elevation['data'][i_ray_psel[ind_vec]]
    coloc_dict['rad2_azi'] = radar2.azimuth['data'][i_ray_psel[ind_vec]]
    coloc_dict['rad2_rng'] = radar2.range['data'][i_rng_psel[ind_vec]]
    coloc_dict['rad2_ray_ind'] = i_ray_psel[ind_vec]
    coloc_dict['rad2_rng_ind'] = i_rng_psel[ind_vec]

    # not colocated: set co-located flag to 1
    coloc_rad1['data'][
        ind_ray_rad1[ind_not_col], ind_rng_rad1[ind_not_col]] = 1

    return coloc_dict, coloc_rad1


def intersection(radar1, radar2, h_tol=0., latlon_tol=0., vol_d_tol=None,
                 vismin=None, hmin=None, hmax=None, rmin=None, rmax=None,
                 elmin=None, elmax=None, azmin=None, azmax=None,
                 visib_field=None, intersec_field=None):
    """
    Flags region of radar1 that is intersecting with radar2 and complies with
    criteria regarding visibility, altitude, range, elevation angle and
    azimuth angle

    Parameters
    ----------
    radar1 : Radar
        radar object that is going to be flagged
    radar2 : Radar
        radar object checked for intersecting region
    h_tol : float
        tolerance in altitude [m]
    latlon_tol : float
        latitude and longitude tolerance [decimal deg]
    vol_d_tol : float
        pulse volume diameter tolerance [m]
    vismin : float
        minimum visibility [percentage]
    hmin, hmax : floats
        min and max altitude [m MSL]
    rmin, rmax : floats
        min and max range from radar [m]
    elmin, elmax : floats
        min and max elevation angle [deg]
    azmin, azmax : floats
        min and max azimuth angle [deg]

    Returns
    -------
    intersec_rad1_dict : dict
        the field with the gates of radar1 in the same region as radar2
        flagged, i.e.: 1 not intersecting, 2 intersecting, 0 is reserved

    """
    # parse the field parameters
    if intersec_field is None:
        intersec_field = get_field_name('colocated_gates')
    if visib_field is None:
        visib_field = get_field_name('visibility')

    # define common volume
    intersec_rad1 = find_intersection_volume(
        radar1, radar2, h_tol=h_tol, latlon_tol=latlon_tol)

    # check for equal volume of rad1
    if vol_d_tol is not None:
        intersec_rad1[np.logical_not(find_equal_vol_region(
            radar1, radar2, vol_d_tol=vol_d_tol))] = 1

    # check for visibility
    if visib_field in radar1.fields and vismin is not None:
        intersec_rad1[radar1.fields[visib_field]['data'] < vismin] = 1

    # check for altitude limits
    if hmin is not None:
        intersec_rad1[radar1.gate_altitude['data'] < hmin] = 1
    if hmax is not None:
        intersec_rad1[radar1.gate_altitude['data'] > hmax] = 1

    # check for range limits
    if rmin is not None:
        intersec_rad1[:, radar1.range['data'] < rmin] = 1
    if rmax is not None:
        intersec_rad1[:, radar1.range['data'] > rmax] = 1

    # check elevation angle limits
    if elmin is not None:
        intersec_rad1[radar1.elevation['data'] < elmin, :] = 1
    if elmax is not None:
        intersec_rad1[radar1.elevation['data'] > elmax, :] = 1

    # check min and max azimuth angle
    if azmin is not None and azmax is not None:
        if azmin <= azmax:
            intersec_rad1[radar1.azimuth['data'] < azmin, :] = 1
            intersec_rad1[radar1.azimuth['data'] > azmax, :] = 1
        if azmin > azmax:
            intersec_rad1[np.logical_and(
                radar1.azimuth['data'] < azmin,
                radar1.azimuth['data'] > azmax), :] = 1
    elif azmin is not None:
        intersec_rad1[radar1.azimuth['data'] < azmin, :] = 1
    elif azmax is not None:
        intersec_rad1[radar1.azimuth['data'] > azmax, :] = 1

    intersec_rad1_dict = get_metadata(intersec_field)
    intersec_rad1_dict['data'] = intersec_rad1
    intersec_rad1_dict.update({'_FillValue': 0})

    return intersec_rad1_dict


def find_intersection_volume(radar1, radar2, h_tol=0., latlon_tol=0.):
    """
    Flags region of radar1 that is intersecting with radar2

    Parameters
    ----------
    radar1 : Radar
        radar object that is going to be flagged
    radar2 : Radar
        radar object checked for intersecting region
    h_tol : float
        tolerance in altitude [m]
    latlon_tol : float
        latitude and longitude tolerance [decimal deg]

    Returns
    -------
    intersec : 2d array
        the field with gates within the common volume flagged, i.e.
        1: Not intersecting, 2: intersecting (0 is reserved)

    """
    intersec = np.ma.ones((radar1.nrays, radar1.ngates), dtype=np.uint8)

    min_lat, max_lat, min_lon, max_lon, min_alt, max_alt = (
        find_intersection_limits(
            radar1.gate_latitude['data'], radar1.gate_longitude['data'],
            radar1.gate_altitude['data'], radar2.gate_latitude['data'],
            radar2.gate_longitude['data'], radar2.gate_altitude['data'],
            h_tol=h_tol, latlon_tol=latlon_tol))

    intersec[np.logical_and(
        np.logical_and(radar1.gate_altitude['data'] > min_alt,
                       radar1.gate_altitude['data'] < max_alt),
        np.logical_and(
            np.logical_and(radar1.gate_latitude['data'] > min_lat,
                           radar1.gate_latitude['data'] < max_lat),
            np.logical_and(radar1.gate_longitude['data'] > min_lon,
                           radar1.gate_longitude['data'] < max_lon)))] = 2

    return intersec


def find_intersection_limits(lat1, lon1, alt1, lat2, lon2, alt2, h_tol=0.,
                             latlon_tol=0.):
    """
    Find the limits of the intersection between two volumes

    Parameters
    ----------
    lat1, lon1, alt1 : float array
        array with the positions of first volume. lat, lon in decimal
        degrees, alt in m MSL.
    lat2, lon2, alt2 : float array
        array with the positions of second volume. lat, lon in decimal
        degrees, alt in m MSL.
    h_tol: float
        altitude tolerance [m MSL]
    latlon_tol: float
        latitude and longitude tolerance [decimal deg]

    Returns
    -------
    min_lat, max_lat, min_lon, max_lon, min_alt, max_alt : floats
        the limits of the intersecting region

    """

    min_lat = np.max([np.min(lat1), np.min(lat2)])-latlon_tol
    max_lat = np.min([np.max(lat1), np.max(lat2)])+latlon_tol

    min_lon = np.max([np.min(lon1), np.min(lon2)])-latlon_tol
    max_lon = np.min([np.max(lon1), np.max(lon2)])+latlon_tol

    min_alt = np.max([np.min(alt1), np.min(alt2)])-h_tol
    max_alt = np.min([np.max(alt1), np.max(alt2)])+h_tol

    return min_lat, max_lat, min_lon, max_lon, min_alt, max_alt


def find_equal_vol_region(radar1, radar2, vol_d_tol=0):
    """
    Flags regions of radar1 that are equivolumetric
    (similar pulse volume diameter) with radar2

    Parameters
    ----------
    radar1 : Radar
        radar object that is going to be flagged
    radar2 : Radar
        radar object
    vol_d_tol : float
        pulse volume diameter tolerance

    Returns
    -------
    equal_vol : 2D boolean array
        field with true where both radars have a similar pulse volume diameter

    """
    rng_ground = get_ground_distance(
        radar1.gate_latitude['data'], radar1.gate_longitude['data'],
        radar2.latitude['data'], radar2.longitude['data'])
    rng_rad2 = get_range(
        rng_ground, radar1.gate_altitude['data'], radar2.altitude['data'])
    if (radar2.instrument_parameters is not None and
            'radar_beam_width_h' in radar2.instrument_parameters):
        bwidth2 = radar2.instrument_parameters['radar_beam_width_h']['data'][0]
    else:
        warn('Unknown radar 2 beamwidth. Assumed 1 deg')
        bwidth2 = 1.
    if (radar1.instrument_parameters is not None and
            'radar_beam_width_h' in radar1.instrument_parameters):
        bwidth1 = radar1.instrument_parameters['radar_beam_width_h']['data'][0]
    else:
        warn('Unknown radar 1 beamwidth. Assumed 1 deg')
        bwidth1 = 1.
    vol_d_rad2 = get_vol_diameter(bwidth2, rng_rad2)
    vol_d_rad1 = get_vol_diameter(
        bwidth1,
        np.broadcast_to(
            radar1.range['data'].reshape(1, radar1.ngates),
            (radar1.nrays, radar1.ngates)))

    return np.isclose(vol_d_rad1, vol_d_rad2, rtol=0., atol=vol_d_tol)


def get_ground_distance(lat_array, lon_array, lat0, lon0):
    """
    Computes the ground distance to a fixed point

    Parameters
    ----------
    lat_array : float array
        array of latitudes [decimal deg]
    lon_array : float array
        array of longitudes [decimal deg]
    lat0: float
        latitude of fix point
    lon0: float
        longitude of fix point

    Returns
    -------
    rng_ground : float array
        the ground range [m]

    """
    # distance of each gate of rad1 from rad2
    r_earth = 6371e3  # [m]

    dlat_rad = (lat_array-lat0)*np.pi/180.
    dlon_rad = (lon_array-lon0)*np.pi/180.
    a = (np.sin(dlat_rad/2.)*np.sin(dlat_rad/2.) +
         np.cos(lat_array*np.pi/180.)*np.cos(lat0*np.pi/180.) *
         np.sin(dlon_rad/2.)*np.sin(dlon_rad/2.))

    return 2.*np.arctan2(np.sqrt(a), np.sqrt(1.-a))*r_earth


def get_range(rng_ground, alt_array, alt0):
    """
    Computes the range to a fixed point from the ground distance and the
    altitudes

    Parameters
    ----------
    rng_ground : float array
        array of ground distances [m]
    alt_array : float array
        array of altitudes [m MSL]
    alt0: float
        altitude of fixed point [m MSL]

    Returns
    -------
    rng : float array
        the range [m]

    """
    alt_from0 = np.abs(alt_array-alt0)

    return np.sqrt(alt_from0*alt_from0+rng_ground*rng_ground)


def get_vol_diameter(beamwidth, rng):
    """
    Computes the pulse volume diameter from the antenna beamwidth and the
    range from the radar

    Parameters
    ----------
    beamwidth : float
        the radar beamwidth [deg]
    rng : float array
        the range from the radar [m]

    Returns
    -------
    vol_d : float array
        the pulse volume diameter

    """

    return beamwidth*np.pi/180.*rng


def _construct_xsect_radar(
        radar, scan_type, pxsect_rays, xsect_sweep_start_ray_index,
        xsect_sweep_end_ray_index, target_angles):
    """
    Constructs a new radar object that contains cross-sections at fixed angles
    of a PPI or RHI volume scan.

    Parameters
    ----------
    radar : Radar
        Radar volume containing RHI/PPI sweeps from which a cross sections will
        be extracted.
    scan_type : str
        Type of cross section scan (ppi or rhi).
    pxsect_rays : list
        list of rays from the radar volume to be copied in the cross-sections
        radar object
    xsect_sweep_start_ray_index, xsect_sweep_end_ray_index : array of ints
        start and end sweep ray index of each cross-section scan
    target_angles : array
        the target fixed angles

    Returns
    -------
    radar_xsect : Radar
        Radar volume containing sweeps which contain cross sections from the
        original volume.

    """
    xsect_nsweeps = len(target_angles)

    _range = _copy_dic(radar.range)
    latitude = _copy_dic(radar.latitude)
    longitude = _copy_dic(radar.longitude)
    altitude = _copy_dic(radar.altitude)
    metadata = _copy_dic(radar.metadata)

    time = _copy_dic(radar.time, excluded_keys=['data'])
    time['data'] = radar.time['data'][pxsect_rays].copy()

    azimuth = _copy_dic(radar.azimuth, excluded_keys=['data'])
    azimuth['data'] = radar.azimuth['data'][pxsect_rays].copy()

    elevation = _copy_dic(radar.elevation, excluded_keys=['data'])
    elevation['data'] = radar.elevation['data'][pxsect_rays].copy()

    fields = {}
    for field_name, orig_field_dic in radar.fields.items():
        field_dic = _copy_dic(orig_field_dic, excluded_keys=['data'])
        field_dic['data'] = orig_field_dic['data'][pxsect_rays].copy()
        fields[field_name] = field_dic

    sweep_number = _copy_dic(radar.sweep_number, excluded_keys=['data'])
    sweep_number['data'] = np.arange(xsect_nsweeps, dtype='int32')

    sweep_mode = _copy_dic(radar.sweep_mode, excluded_keys=['data'])
    sweep_mode['data'] = np.array([scan_type]*xsect_nsweeps)

    fixed_angle = _copy_dic(radar.fixed_angle, excluded_keys=['data'])
    fixed_angle['data'] = np.array(target_angles, dtype='float32')

    sweep_start_ray_index = _copy_dic(
        radar.sweep_start_ray_index, excluded_keys=['data'])
    ssri = np.array(xsect_sweep_start_ray_index, dtype='int32')
    sweep_start_ray_index['data'] = ssri

    sweep_end_ray_index = _copy_dic(
        radar.sweep_end_ray_index, excluded_keys=['data'])
    seri = np.array(xsect_sweep_end_ray_index, dtype='int32')
    sweep_end_ray_index['data'] = seri

    radar_xsect = Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle,
        sweep_start_ray_index, sweep_end_ray_index,
        azimuth, elevation)

    return radar_xsect


def _copy_dic(orig_dic, excluded_keys=None):
    """ Return a copy of the original dictionary copying each element. """
    if excluded_keys is None:
        excluded_keys = []
    dic = {}
    for k, v in orig_dic.items():
        if k not in excluded_keys:
            dic[k] = copy(v)
    return dic



"""
pyart.aux_io.rainbow_psr
========================

Routines for reading RAINBOW PSR files (Used by SELEX)

.. autosummary::
    :toctree: generated/

    read_iq
    read_iq_data

"""

# specific modules for this function
import os
from warnings import warn
from copy import deepcopy

import numpy as np
from scipy.constants import speed_of_light

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.radar_spectra import RadarSpectra
from ..util import cut_radar
from .metranet_reader import read_metranet
from .rainbow_psr import get_Doppler_info

IQ_FIELD_NAMES = {
    'IQhhADU': 'IQ_hh_ADU',
    'IQvvADU': 'IQ_vv_ADU',
}


def read_iq(filename, filenames_iq, field_names=None,
            additional_metadata=None, file_field_names=False,
            exclude_fields=None, include_fields=None, reader='C', nbytes=4,
            prf=None, ang_tol=0.4, noise_h=None, noise_v=None, rconst_h=None,
            rconst_v=None, radconst_h=None, radconst_v=None, mfloss_h=1.,
            mfloss_v=1., azi_min=None, azi_max=None, ele_min=None,
            ele_max=None, rng_min=None, rng_max=None, **kwargs):
    """
    Read a rad4alp IQ file.

    Parameters
    ----------
    filename : str
        Name of the METRANET file to be used as reference.
    filenames_iq : list of str
        Name of the IQ files
    field_names : dict, optional
        Dictionary mapping RAINBOW field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    reader : str
        The library used to read the METRANET reference file. Can be either
        'C' or 'python'
    nbytes : int
        The number of bytes used to store the data in numpy arrays, e.g. if
        nbytes=4 then floats are going to be stored as np.float32
    prf : float
        The PRF of the read scan
    ang_tol : float
        Tolerated angle distance between nominal radar angle and angle in
        PSR files
    noise_h, noise_v : float
        The estimated H(V) noise power (ADU) of the scan
    rconst_h, rconst_v : float
        Dynamical factor used in the conversion from dBADU to dBm/dBZ
    radconst_h, radconst_v : float
        The H(V) radar constant
    mfloss_h, mfloss_v : float
        The H(V) matched filter losses in the receiver (dB)
    azi_min, azi_max, ele_min, ele_max : float or None
        The minimum and maximum angles to keep (deg)
    rng_min, rng_max : float or None
        The minimum and maximum ranges to keep (m)

    Returns
    -------
    radar : Radar
        Radar object containing data from PSR file.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create radar object used as reference
    try:
        radar = read_metranet(
            filename, field_names={'ZH': 'reflectivity'}, reader=reader,
            nbytes=nbytes)
        radar.fields = dict()
        rng_orig = radar.range['data']
        radar = cut_radar(
            radar, None, rng_min=None, rng_max=None, ele_min=ele_min,
            ele_max=ele_max, azi_min=azi_min, azi_max=azi_max)
        if radar is None:
            warn('No data within specified azimuth, elevation and'
                 ' range limits')
            return None
        ind_rng_start = np.where(rng_orig == radar.range['data'][0])[0]
        ind_rng_end = np.where(rng_orig == radar.range['data'][-1])[0]
        ind_rng = np.arange(ind_rng_start, ind_rng_end+1, dtype=int)
    except OSError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None

    # create metadata retrieval object
    if field_names is None:
        field_names = IQ_FIELD_NAMES
    filemetadata = FileMetadata('IQ', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)
    npulses = filemetadata('number_of_pulses')
    dBADU_to_dBm_hh = filemetadata('dBADU_to_dBm_hh')
    dBADU_to_dBm_vv = filemetadata('dBADU_to_dBm_vv')
    mfloss_h_dict = filemetadata('matched_filter_loss_h')
    mfloss_v_dict = filemetadata('matched_filter_loss_v')
    prt = filemetadata('prt')

    # Keep only IQ files with data from radar object
    filenames_iq_aux, npulses['data'] = get_valid_rays(
        filenames_iq, radar.azimuth['data'], radar.fixed_angle['data'][0],
        ang_tol=ang_tol)

    if filenames_iq_aux is None:
        return None

    # Select valid range gates
    if rng_min is None:
        rng_min = 0.
    if rng_max is None:
        rng_max = np.max(radar.range['data'])

    ind_rng = np.where(np.logical_and(
        radar.range['data'] >= rng_min,
        radar.range['data'] <= rng_max))[0]

    if ind_rng.size == 0:
        warn('No range bins between '+str(rng_min)+' and '+str(rng_max)+' m')
        return None

    if ind_rng.size == radar.ngates:
        ind_rng = None

    # Read IQ data
    if 'IQ_hh_ADU' in field_names or 'IQ_vv_ADU' in field_names:
        npulses_max = np.max(npulses['data'])
        data_hh = np.ma.masked_all(
            (radar.nrays, radar.ngates, npulses_max), dtype=np.complex64)
        data_vv = np.ma.masked_all(
            (radar.nrays, radar.ngates, npulses_max), dtype=np.complex64)
        for i, npuls in enumerate(npulses['data']):
            data_ray_hh, data_ray_vv = read_iq_data(
                filenames_iq_aux[i], radar.ngates, npuls)
            if data_ray_hh is not None:
                data_hh[i, :, 0:npuls] = data_ray_hh
            if data_ray_vv is not None:
                data_vv[i, :, 0:npuls] = data_ray_vv

        if ind_rng is not None:
            data_hh = data_hh[:, ind_rng, :]
            data_vv = data_vv[:, ind_rng, :]

    # cut radar range
    if ind_rng is not None:
        radar.range['data'] = radar.range['data'][ind_rng]
        radar.init_gate_x_y_z()
        radar.init_gate_longitude_latitude()
        radar.init_gate_altitude()
        radar.ngates = ind_rng.size

    # Prepare output fields
    fields = {}
    for field_name in field_names:
        field_dict = filemetadata(field_name)
        if field_name == 'IQ_hh_ADU':
            field_dict['data'] = data_hh
        elif field_name == 'IQ_vv_ADU':
            field_dict['data'] = data_vv
        elif field_name == 'IQ_noise_power_hh_ADU' and noise_h is not None:
            npulses_max = np.max(npulses['data'])
            field_dict['data'] = np.ma.masked_all(
                (radar.nrays, radar.ngates, npulses_max), dtype=np.float)
            for i, npuls in enumerate(npulses['data']):
                field_dict['data'][i, :, 0:npuls] = noise_h
        elif field_name == 'IQ_noise_power_vv_ADU' and noise_v is not None:
            npulses_max = np.max(npulses['data'])
            field_dict['data'] = np.ma.masked_all(
                (radar.nrays, radar.ngates, npulses_max), dtype=np.float)
            for i, npuls in enumerate(npulses['data']):
                field_dict['data'][i, :, 0:npuls] = noise_v
        else:
            warn('Field name '+field_name+' not known')
            continue
        fields[field_name] = field_dict

    # get further metadata
    if rconst_h is not None and radconst_h is not None and mfloss_h is not None:
        dBADU_to_dBm_hh['data'] = np.array([rconst_h-40.-radconst_h-mfloss_h])
        if radar.radar_calibration is None:
            radar.radar_calibration = {'dBADU_to_dBm_hh': dBADU_to_dBm_hh}
        else:
            radar.radar_calibration.update(
                {'dBADU_to_dBm_hh': dBADU_to_dBm_hh})
    else:
        warn('Unable to compute dBADU_to_dBm_hh. Missing data')

    if rconst_v is not None and radconst_v is not None and mfloss_v is not None:
        dBADU_to_dBm_vv['data'] = np.array([rconst_v-40.-radconst_v-mfloss_v])
        if radar.radar_calibration is None:
            radar.radar_calibration = {'dBADU_to_dBm_vv': dBADU_to_dBm_vv}
        else:
            radar.radar_calibration.update(
                {'dBADU_to_dBm_vv': dBADU_to_dBm_vv})
    else:
        warn('Unable to compute dBADU_to_dBm_vv. Missing data')

    if mfloss_h is not None:
        mfloss_h_dict['data'] = np.array([mfloss_h])
        if radar.radar_calibration is None:
            radar.radar_calibration = {'matched_filter_loss_h': mfloss_h_dict}
        else:
            radar.radar_calibration.update(
                {'matched_filter_loss_h': mfloss_h_dict})
    else:
        warn('matched_filter_loss_h not known')

    if mfloss_v is not None:
        mfloss_v_dict['data'] = np.array([mfloss_v])
        if radar.radar_calibration is None:
            radar.radar_calibration = {'matched_filter_loss_v': mfloss_v_dict}
        else:
            radar.radar_calibration.update(
                {'matched_filter_loss_v': mfloss_v_dict})
    else:
        warn('matched_filter_loss_v not known')

    if prf is not None:
        prt['data'] = np.zeros(radar.nrays)+1./prf
        if radar.instrument_parameters is None:
            radar.instrument_parameters = {'prt': prt}
        else:
            radar.instrument_parameters.update({'prt': prt})
    else:
        warn('prt not known')

    Doppler_velocity = None
    Doppler_frequency = None
    if (prf is not None and radar.instrument_parameters is not None and
            'frequency' in radar.instrument_parameters):
        Doppler_velocity = filemetadata('Doppler_velocity')
        Doppler_frequency = filemetadata('Doppler_frequency')
        vel_data, freq_data = get_Doppler_info(
            np.zeros(radar.nrays)+prf, npulses['data'],
            speed_of_light/radar.instrument_parameters['frequency']['data'][0],
            fold=True)
        Doppler_velocity['data'] = vel_data
        Doppler_frequency['data'] = freq_data

    return RadarSpectra(
        radar.time, radar.range, fields, radar.metadata,
        radar.scan_type, radar.latitude, radar.longitude, radar.altitude,
        radar.sweep_number, radar.sweep_mode, radar.fixed_angle,
        radar.sweep_start_ray_index, radar.sweep_end_ray_index,
        radar.azimuth, radar.elevation, npulses,
        Doppler_velocity=Doppler_velocity,
        Doppler_frequency=Doppler_frequency,
        rays_are_indexed=radar.rays_are_indexed,
        ray_angle_res=radar.ray_angle_res,
        instrument_parameters=radar.instrument_parameters,
        radar_calibration=radar.radar_calibration)


def read_iq_data(filename, ngates, npulses, nchannels=2):
    """
    Reads the IQ data

    Parameters
    ----------
    filename : str
        Name of file containing the IQ data of a ray
    ngates : int
        Number of gates in ray
    npulses : int
        Number of pulses in ray
    nchannels : int
        Number of channels in file

    Returns
    -------
    data_hh, data_vv : 2D array
        arrays containing the HH and VV channels IQ data

    """
    try:
        with open(filename, 'rb') as file:
            # file.readline()
            data = np.fromfile(
                file, dtype=np.complex64,
                count=ngates*npulses*nchannels)
            if data.size != ngates*npulses*nchannels:
                warn('Data file containing '+str(data.size)+' elements. ' +
                     str(ngates*npulses*nchannels)+' expected.')
                return None, None

            data = np.reshape(data, [ngates, nchannels, npulses], order='F')
            data_hh = data[:, 0, :]
            data_vv = data[:, 1, :]

            return data_hh, data_vv

    except EnvironmentError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None, None


def get_valid_rays(filenames_iq, ref_azi, ref_ele, ang_tol=0.4):
    """
    Selects the IQ files corresponding to each ray azimuth and gets
    the number of pulses corresponding to each ray

    Parameters
    ----------
    filenames_iq : list of str
        List of files containing the IQ information of each ray
    ref_azi : float array
        The radar azimuths (deg)
    ref_ele : int
        The elevation of the current scan (deg)
    ang_tol : float
        The angle tolerance [deg] between the reference azimuth angle of the
        radar object and that of the IQ file

    Returns
    -------
    filenames_iq_out : array of strings
        List of files containing IQ information for valid rays
    npulses_vec : array of ints
        Number of pulses for each ray

    """
    # Get azimuth from file name
    azi_vec = []
    for filename_iq in filenames_iq:
        azi_vec.append(float(os.path.basename(filename_iq)[39:42])+0.5)
    azi_vec = np.array(azi_vec)
    filenames_aux = np.array(deepcopy(filenames_iq))

    # Keep only IQ files with data from radar object
    filenames_iq_out = []
    npulses_vec = []
    for azi in ref_azi:
        ind = np.where(np.logical_and(
            azi_vec >= azi-ang_tol, azi_vec <= azi+ang_tol))[0]

        if ind.size == 0:
            warn('No file found for azimuth angle '+str(azi))
            return None, None
        if ind.size > 1:
            filenames_aux2 = filenames_aux[ind]
            ele_vec = []
            for fname in filenames_aux2:
                ele_vec.append(float(os.path.basename(fname)[34:38])/100.)

            delta_ele_vec = np.abs(ele_vec-ref_ele)
            ind = ind[np.argmin(delta_ele_vec)]
        else:
            ind = ind[0]

        filenames_iq_out.append(filenames_aux[ind])
        npulses_vec.append(
            int(int(os.path.basename(str(filenames_aux[ind]))[49:54])/2))

    return np.array(filenames_iq_out), np.array(npulses_vec)



"""
pyart.io.cf1_cartesian
======================

Utilities for reading CF1 Cartesian files.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    _NetCDFVariableDataExtractor

.. autosummary::
    :toctree: generated/

    read_cf1_cartesian
    read_cf1_cartesian_mf

"""

from warnings import warn

import numpy as np
import netCDF4

from ..config import FileMetadata, get_metadata
from ..io.common import _test_arguments
from ..core.grid import Grid
from ..io.cfradial import _ncvar_to_dict

SAT_FIELD_NAMES = [
    'IR_016', 'IR_039', 'IR_087', 'IR_097', 'IR_108', 'IR_120', 'IR_134',
    'CTH', 'HRV', 'VIS006', 'VIS008', 'WV_062', 'WV_073']

MF_FIELD_NAMES = ['radar_estimated_rain_rate', 'data_quality']

MF_FIELD_NAMES_DICT = {
    'data_quality': 'QUALITY',
    'radar_estimated_rain_rate': 'ACRR'}


def read_cf1_cartesian(filename, field_names=None, delay_field_loading=False,
                       chy0=255., chx0=-160., **kwargs):
    """
    Read a CF-1 netCDF file.

    Parameters
    ----------
    filename : str
        Name of CF/Radial netCDF file to read data from.
    field_names : list of fields to read from file.
        If None all files will be read.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed.  In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects.  Delayed field loading will not
        provide any speedup in file where the number of gates vary between
        rays (ngates_vary=True) and is not recommended.
    chy0, chx0 : float
        Swiss coordinates position of the south-western point in the grid

    Returns
    -------
    grid : Grid
        Grid object containing data from METRANET file.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    reserved_variables = [
        'band0', 'band1', 'band10', 'band11', 'band12', 'band2', 'band3',
        'band4', 'band5', 'band6', 'band7', 'band8', 'band9',
        'grid_mapping_0', 'nominal_wavelength0', 'nominal_wavelength1',
        'nominal_wavelength10', 'nominal_wavelength11',
        'nominal_wavelength12', 'nominal_wavelength2', 'nominal_wavelength3',
        'nominal_wavelength4', 'nominal_wavelength5', 'nominal_wavelength6',
        'nominal_wavelength7', 'nominal_wavelength8', 'nominal_wavelength9',
        'time', 'wl_CTH', 'wl_HRV', 'wl_IR_016', 'wl_IR_039', 'wl_IR_087',
        'wl_IR_097', 'wl_IR_108', 'wl_IR_120', 'wl_IR_134', 'wl_VIS006',
        'wl_VIS008', 'wl_WV_062', 'wl_WV_073', 'x0', 'y0']

    # create metadata retrieval object
    if field_names is None:
        field_names = SAT_FIELD_NAMES
    filemetadata = FileMetadata('satellite')

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables
    ncdims = ncobj.dimensions

    nx = ncdims['x0'].size
    ny = ncdims['y0'].size

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])

    grid_mapping = _ncvar_to_dict(ncvars['grid_mapping_0'])

    x['data'] = (
        1000.*(np.arange(nx)+chy0+1/2.)-grid_mapping['false_easting'])

    y['data'] = (
        1000.*(np.arange(ny)+chx0+1/2.)-grid_mapping['false_northing'])

    z['data'] = np.array([0.])

    # Origin of LV03 Swiss coordinates
    origin_latitude['data'] = np.array(
        [grid_mapping['latitude_of_projection_origin']])
    origin_longitude['data'] = np.array(
        [grid_mapping['longitude_of_projection_origin']])
    origin_altitude['data'] = np.array([0.])

    time = _ncvar_to_dict(ncvars['time'])

    # projection (Swiss Oblique Mercator)
    projection = {
        'proj': 'somerc',
        '_include_lon_0_lat_0': True
    }

    # read in the fields
    fields = {}

    # fields in the file has a shape of (ny, nx, 1) with 1
    # indicating frequency band
    # but should shaped (nz, ny, nx) in the Grid object
    field_shape = (ny, nx, 1)

    # check all non-reserved variables, those with the correct shape
    # are added to the field dictionary, if a wrong sized field is
    # detected a warning is raised
    field_keys = [k for k in ncvars if k not in reserved_variables]
    for field in field_names:
        if field not in field_keys:
            warn('Field '+field+' not in file')
            continue
        field_dic_file = _ncvar_to_dict(ncvars[field])
        if field_dic_file['data'].shape == field_shape:
            field_dic_file['data'] = np.expand_dims(
                np.squeeze(field_dic_file['data'], axis=-1),
                axis=0)[:, ::-1, :]
            field_dic = get_metadata(field)  # get field definition from Py-ART
            field_dic['data'] = field_dic_file['data']
            fields[field] = field_dic
        else:
            bad_shape = field_dic_file['data'].shape
            warn(
                'Field %s skipped due to incorrect shape %s'
                % (field, bad_shape))

    # radar_ variables
    if 'radar_latitude' in ncvars:
        radar_latitude = _ncvar_to_dict(ncvars['radar_latitude'])
    else:
        radar_latitude = None

    if 'radar_longitude' in ncvars:
        radar_longitude = _ncvar_to_dict(ncvars['radar_longitude'])
    else:
        radar_longitude = None

    if 'radar_altitude' in ncvars:
        radar_altitude = _ncvar_to_dict(ncvars['radar_altitude'])
    else:
        radar_altitude = None

    if 'radar_name' in ncvars:
        radar_name = _ncvar_to_dict(ncvars['radar_name'])
    else:
        radar_name = None

    if 'radar_time' in ncvars:
        radar_time = _ncvar_to_dict(ncvars['radar_time'])
    else:
        radar_time = None

    # do not close file if field loading is delayed
    if not delay_field_loading:
        ncobj.close()

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)


def read_cf1_cartesian_mf(filename, field_names=None,
                          delay_field_loading=False, **kwargs):
    """
    Read a CF-1 netCDF file.

    Parameters
    ----------
    filename : str
        Name of CF/Radial netCDF file to read data from.
    field_names : list of fields to read from file.
        If None all files will be read.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed.  In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects.  Delayed field loading will not
        provide any speedup in file where the number of gates vary between
        rays (ngates_vary=True) and is not recommended.
    chy0, chx0 : float
        Swiss coordinates position of the south-western point in the grid

    Returns
    -------
    grid : Grid
        Grid object containing data from METRANET file.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    reserved_variables = [
        'X', 'Y', 'time', 'grid_mapping', 'time_coverage_start',
        'time_coverage_end', 'data_mf_products', 'radar_image_time',
        'radar_image_latitude', 'radar_image_longitude', 'radar_image_wmoid',
        'radar_image_name']

    # create metadata retrieval object
    if field_names is None:
        field_names = MF_FIELD_NAMES
    filemetadata = FileMetadata('MF')

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables
    ncdims = ncobj.dimensions

    nx = ncdims['X'].size
    ny = ncdims['Y'].size

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])

    grid_mapping = _ncvar_to_dict(ncvars['grid_mapping'])

    x['data'] = _ncvar_to_dict(ncvars['X'])['data']
    y['data'] = _ncvar_to_dict(ncvars['Y'])['data']
    z['data'] = np.array([0.])

    # Origin
    origin_latitude['data'] = np.array(
        [grid_mapping['latitude_of_projection_origin']])
    origin_longitude['data'] = np.array(
        [grid_mapping['straight_vertical_longitude_from_pole']])
    origin_altitude['data'] = np.array([0.])

    time = _ncvar_to_dict(ncvars['time'])

    # projection
    projection = metadata['crs_proj4_string']

    # read in the fields
    fields = {}

    field_shape = (1, nx, ny)

    # check all non-reserved variables, those with the correct shape
    # are added to the field dictionary, if a wrong sized field is
    # detected a warning is raised
    field_keys = [k for k in ncvars if k not in reserved_variables]
    for field in field_names:
        if MF_FIELD_NAMES_DICT[field] not in field_keys:
            warn('Field '+field+' not in file')
            continue
        field_dic_file = _ncvar_to_dict(ncvars[MF_FIELD_NAMES_DICT[field]])
        if field_dic_file['data'].shape == field_shape:
            field_dic_file['data'] = np.ma.transpose(
                field_dic_file['data'], axes=[0, 2, 1])
            if field == 'radar_estimated_rain_rate':
                # put 1/100 mm/ 5 min in mm/h
                field_dic_file['data'] = 0.12*field_dic_file['data']
            field_dic = get_metadata(field)  # get field definition from Py-ART
            field_dic['data'] = field_dic_file['data']
            fields[field] = field_dic
        else:
            bad_shape = field_dic_file['data'].shape
            warn(
                'Field %s skipped due to incorrect shape %s'
                % (field, bad_shape))

    # radar_ variables
    if 'radar_image_latitude' in ncvars:
        radar_latitude = _ncvar_to_dict(ncvars['radar_image_latitude'])
    else:
        radar_latitude = None

    if 'radar_image_longitude' in ncvars:
        radar_longitude = _ncvar_to_dict(ncvars['radar_image_longitude'])
    else:
        radar_longitude = None

    if 'radar_altitude' in ncvars:
        radar_altitude = _ncvar_to_dict(ncvars['radar_altitude'])
    else:
        radar_altitude = None

    if 'radar_image_name' in ncvars:
        radar_name = _ncvar_to_dict(ncvars['radar_image_name'])
    else:
        radar_name = None

    if 'radar_image_time' in ncvars:
        radar_time = _ncvar_to_dict(ncvars['radar_image_time'])
    else:
        radar_time = None

    # do not close file if field loading is delayed
    if not delay_field_loading:
        ncobj.close()

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)



#! /usr/bin/env python
"""
Make a small sigmet (IRIS) file containing a single PPI scan.

Reflectivity data is set to 0 dBZ, other parameters are taken from the
full sized file XSW110520105408.RAW7HHF
"""

import struct
import numpy as np

# parameters
INFILE = 'XSW110520105408.RAW7HHF'
OUTFILE = 'example_sigmet_ppi.sigmet'

# open the input and output files
f = open(INFILE, 'rb')
out = open(OUTFILE, 'wb')

#############################
# First Record (6144 bytes) #
# product_hdr (640 bytes)   #
#############################

# structure_header (12 bytes)
out.write(f.read(12))       # no change

# product_configuration (320 bytes)
# -> structure_header
# -> generation_time
# -> sweep_ingest_time
# -> file_ingest_time
# -> color_scale_def
out.write(f.read(320))       # no change

# product_end (308 bytes)
# -> ingest_time
fmt = ('16s8s8s12s28sh16s16shIIhhiiHHh12sHiiiiiHhHhhhHhhhhHH18sIIIIIIHHIIh' +
       '32shhhBB2s8sI4s')
l = list(struct.unpack(fmt, f.read(308)))
l[23] = 154000      # last_bin_range
l[24] = 25          # number_bins
out.write(struct.pack(fmt, *l))

# padding to next record
out.write(f.read(6144 - 640))       # padding to next record

##############################
# Second Record (6144 bytes) #
# ingest_header (4884 bytes) #
##############################
# structure_header
out.write(f.read(12))       # no change

# ingest_configuration (480 bytes)
# -> volume_scan_start_time
# ->
fmt = '80shhi12s12shhhh4s8s16sh16shIIhhHHHhiiiiiiiIh2s8sI16s228s'
l = list(struct.unpack(fmt, f.read(480)))
l[2] = 1        # number_sweeps_completed
l[22] = 20      # number_rays_sweep
out.write(struct.pack(fmt, *l))

# task_configuration (2612 bytes)
# -> structure_header (12 byes)
out.write(f.read(12))               # no change

# -> task_sched_info (120 bytes)
out.write(f.read(120))              # no change

# -------------------------------------------------
# -> task_dsp_info (320 bytes)
# ---> major_mode, dsp_type (4 bytes)
out.write(f.read(4))

# ---> current_data_type_mask (24 bytes)
fmt = 'IIIIII'
l = list(struct.unpack(fmt, f.read(24)))
l[0] = 512      # mask_word_0 indicating field 9 (DBZ2), 2 ** 9
l[2] = 0        # mask_word_1 indicating no fields
out.write(struct.pack(fmt, *l))

# ---> original_data_type_mask, task_dsp_mode, etc (292 bytes)
out.write(f.read(292))
# ---------------------------------------------------

# -> task_calib_info (320 bytes)
out.write(f.read(320))              # no change

# -> task_range_info (160 bytes) - file at 7408 seek
fmt = 'iihhiiHh136s'
l = list(struct.unpack(fmt, f.read(160)))
l[1] = 154000       # last_bin_range
l[2] = 25           # number_input_bins
l[3] = 25           # number_output_bins
out.write(struct.pack(fmt, *l))

# -> task_scan_info (320 bytes)
fmt = 'Hh2sh200s112s'
l = list(struct.unpack(fmt, f.read(320)))
l[3] = 1    # number_sweeps
out.write(struct.pack(fmt, *l))

# -> task_misc_info (320 bytes)
out.write(f.read(320))              # no change

# -> task_end_info (320 bytes)
out.write(f.read(320))              # no change

# -> comments (720 bytes)
out.write(f.read(720))              # no change


# spare_0 (732 bytes)
out.write(f.read(732))      # no change

# gparm (128 bytes)
out.write(f.read(128))      # no change

# reserved (920 bytes)
out.write(f.read(920))      # no change

# padding to next record
out.write(f.read(6144-4884))

###################
# Data            #
###################
# f is at 12288
# raw_prod_bhdr (12 bytes)
out.write(f.read(12))

# skip over ingest_data_header for field 8
f.read(76)

# ingest_data_header for field 9
fmt = '12s12shhhhhHhH36s'
l = list(struct.unpack(fmt, f.read(76)))
l[3] = 20       # number_rays_sweep
l[5] = 20       # number_rays_file_expected
l[6] = 20       # number_rays_file_actual
out.write(struct.pack(fmt, *l))

# compressed ray data
code = np.ones((33), dtype='int16')
code[0] = -32737        # indicates 31 values follow (-32737 + 32768 = 31)
code[2] = 91            # el0
code[4] = 91            # el1
code[5] = 25            # nbins
code[6:-1] = -32768     # 0 dBZ
code[-1] = 1            # end of ray marker

for i in range(19):
    code[6] = i     # time
    code.view('uint16')[1] = 182 * 18 * i   # az0 0 -> 360 in steps of 18
    code.view('uint16')[3] = 182 * 18 * i   # az1
    out.write(code.tostring())

# last ray (20th ray, index 19) has 15 bins instead of 25.
code = np.ones((23), dtype='int16')
code[0] = -32747        # indicates 21 values follow (-32747 + 32768 = 21)
code[2] = 91            # el0
code[4] = 91            # el1
code[5] = 15            # 15 nbins in last ray
code[6:-1] = -32768     # 0 dBZ
code[-1] = 1            # end of ray marker

code[6] = 19     # time
code.view('uint16')[1] = 182 * 18 * 19   # az0 0 -> 360 in steps of 18
code.view('uint16')[3] = 182 * 18 * 19   # az1
out.write(code.tostring())

out.write('\x00' * 4756)

# close the files
f.close()
out.close()



"""
====================================
Create a PPI plot from a Sigmet file
====================================

An example which creates a PPI plot of a Sigmet file.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart

filename = 'XSW110520105408.RAW7HHF'

# create the plot using RadarDisplay (recommended method)
radar = pyart.io.read_rsl(filename)
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure()
ax = fig.add_subplot(111)
display.plot('reflectivity', 0, vmin=-32, vmax=64.)
display.plot_range_rings([10, 20, 30, 40])
display.plot_cross_hair(5.)
plt.show()



"""
pyart.core.transforms
=====================

Transformations between coordinate systems. Routines for converting between
Cartesian/Cartographic (x, y, z), Geographic (latitude, longitude, altitude)
and antenna (azimuth, elevation, range) coordinate systems.

.. autosummary::
    :toctree: generated/

    antenna_to_cartesian
    cartesian_to_antenna
    antenna_vectors_to_cartesian
    antenna_to_cartesian_track_relative
    antenna_to_cartesian_earth_relative
    antenna_to_cartesian_aircraft_relative

    cartesian_to_geographic
    cartesian_vectors_to_geographic
    geographic_to_cartesian
    cartesian_to_geographic_aeqd
    geographic_to_cartesian_aeqd
    swissCH1903_to_wgs84
    wgs84_to_swissCH1903

    _interpolate_axes_edges
    _interpolate_azimuth_edges
    _interpolate_elevation_edges
    _interpolate_range_edges
    _half_angle_complex


"""

import warnings

import numpy as np
try:
    import pyproj
    _PYPROJ_AVAILABLE = True
except ImportError:
    _PYPROJ_AVAILABLE = False

from ..exceptions import MissingOptionalDependency

PI = np.pi


def antenna_to_cartesian(ranges, azimuths, elevations, ke = 4/3.):
    """
    Return Cartesian coordinates from antenna coordinates.

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    azimuths : array
        Azimuth angle of the radar in degrees.
    elevations : array
        Elevation angle of the radar in degrees.
    ke : float, optional
        Effective radius scale factor 

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    The calculation for Cartesian coordinate is adapted from equations
    2.28(b) and 2.28(c) of Doviak and Zrnic [1]_ assuming a
    standard atmosphere (4/3 Earth's radius model).

    .. math::

        z = \\sqrt{r^2+R^2+2*r*R*sin(\\theta_e)} - R

        s = R * arcsin(\\frac{r*cos(\\theta_e)}{R+z})

        x = s * sin(\\theta_a)

        y = s * cos(\\theta_a)

    Where r is the distance from the radar to the center of the gate,
    :math:`\\theta_a` is the azimuth angle, :math:`\\theta_e` is the
    elevation angle, s is the arc length, and R is the effective radius
    of the earth, taken to be 4/3 the mean radius of earth (6371 km).

    References
    ----------
    .. [1] Doviak and Zrnic, Doppler Radar and Weather Observations, Second
        Edition, 1993, p. 21.

    """
    theta_e = elevations * PI / 180.0    # elevation angle in radians.
    theta_a = azimuths * PI / 180.0      # azimuth angle in radians.
    R = 6371.0 * 1000.0 * ke     # effective radius of earth in meters.
    r = ranges * 1000.0                 # distances to gates in meters.

    z = (r ** 2 + R ** 2 + 2.0 * r * R * np.sin(theta_e)) ** 0.5 - R
    s = R * np.arcsin(r * np.cos(theta_e) / (R + z))  # arc length in m.
    x = s * np.sin(theta_a)
    y = s * np.cos(theta_a)
    return x, y, z


def antenna_vectors_to_cartesian(ranges, azimuths, elevations, edges=False,
                                 ke = 4/3.):
    """
    Calculate Cartesian coordinate for gates from antenna coordinate vectors.

    Calculates the Cartesian coordinates for the gate centers or edges for
    all gates from antenna coordinate vectors assuming a standard atmosphere
    (4/3 Earth's radius model). See :py:func:`pyart.util.antenna_to_cartesian`
    for details.

    Parameters
    ----------
    ranges : array, 1D.
        Distances to the center of the radar gates (bins) in meters.
    azimuths : array, 1D.
        Azimuth angles of the rays in degrees.
    elevations : array, 1D.
        Elevation angles of the rays in degrees.
    edges : bool, optional
        True to calculate the coordinates of the gate edges by interpolating
        between gates and extrapolating at the boundaries. False to
        calculate the gate centers.
    ke : float, optional
        Effective radius scale factor 

    Returns
    -------
    x, y, z : array, 2D
        Cartesian coordinates in meters from the center of the radar to the
        gate centers or edges.

    """
    if edges:
        if len(ranges) != 1:
            ranges = _interpolate_range_edges(ranges)
        if len(elevations) != 1:
            elevations = _interpolate_elevation_edges(elevations)
        if len(azimuths) != 1:
            azimuths = _interpolate_azimuth_edges(azimuths)
    rg, azg = np.meshgrid(ranges, azimuths)
    rg, eleg = np.meshgrid(ranges, elevations)
    return antenna_to_cartesian(rg / 1000., azg, eleg, ke)


def _interpolate_range_edges(ranges):
    """ Interpolate the edges of the range gates from their centers. """
    edges = np.empty((ranges.shape[0] + 1, ), dtype=ranges.dtype)
    edges[1:-1] = (ranges[:-1] + ranges[1:]) / 2.
    edges[0] = ranges[0] - (ranges[1] - ranges[0]) / 2.
    edges[-1] = ranges[-1] - (ranges[-2] - ranges[-1]) / 2.
    edges[edges < 0] = 0    # do not allow range to become negative
    return edges


def _interpolate_elevation_edges(elevations):
    """ Interpolate the edges of the elevation angles from their centers. """
    edges = np.empty((elevations.shape[0]+1, ), dtype=elevations.dtype)
    edges[1:-1] = (elevations[:-1] + elevations[1:]) / 2.
    edges[0] = elevations[0] - (elevations[1] - elevations[0]) / 2.
    edges[-1] = elevations[-1] - (elevations[-2] - elevations[-1]) / 2.
    edges[edges > 180] = 180.   # prevent angles from going below horizon
    edges[edges < 0] = 0.
    return edges


def _interpolate_azimuth_edges(azimuths):
    """ Interpolate the edges of the azimuth angles from their centers. """
    edges = np.empty((azimuths.shape[0]+1, ), dtype=azimuths.dtype)
    # perform interpolation and extrapolation in complex plane to
    # account for periodic nature of azimuth angle.
    azimuths = np.exp(1.j*np.deg2rad(azimuths))

    edges[1:-1] = np.angle(azimuths[1:] + azimuths[:-1], deg=True)

    half_angle = _half_angle_complex(azimuths[0], azimuths[1])
    edges[0] = (np.angle(azimuths[0], deg=True) - half_angle) % 360.

    half_angle = _half_angle_complex(azimuths[-1], azimuths[-2])
    edges[-1] = (np.angle(azimuths[-1], deg=True) + half_angle) % 360.

    edges[edges < 0] += 360     # range from [-180, 180] to [0, 360]
    return edges


def _half_angle_complex(complex_angle1, complex_angle2):
    """
    Return half the angle between complex numbers on the unit circle.

    Parameters
    ----------
    complex_angle1, complex_angle2 : complex
        Complex numbers representing unit vectors on the unit circle

    Returns
    -------
    half_angle : float
        Half the angle between the unit vectors in degrees.

    """
    dot_product = np.real(complex_angle1 * np.conj(complex_angle2))
    if dot_product > 1:
        warnings.warn("dot_product is larger than one.")
        dot_product = 1.
    full_angle_rad = np.arccos(dot_product)
    half_angle_rad = full_angle_rad / 2.
    half_angle_deg = np.rad2deg(half_angle_rad)
    return half_angle_deg


def _interpolate_axes_edges(axes):
    """ Interpolate the edges of the axes gates from their centers. """
    edges = np.empty((axes.shape[0] + 1, ), dtype=axes.dtype)
    edges[1:-1] = (axes[:-1] + axes[1:]) / 2.
    edges[0] = axes[0] - (axes[1] - axes[0]) / 2.
    edges[-1] = axes[-1] - (axes[-2] - axes[-1]) / 2.
    return edges


def antenna_to_cartesian_track_relative(ranges, rot, roll, drift, tilt, pitch):
    """
    Calculate track-relative Cartesian coordinates from radar coordinates.

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    roll : array
        Roll angle of the radar in degrees.
    drift : array
        Drift angle of the radar in degrees.
    tilt : array
        Tilt angle of the radar in degrees.
    pitch : array
        Pitch angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    track-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    roll = np.radians(roll)             # roll angle in radians.
    drift = np.radians(drift)           # drift angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    pitch = np.radians(pitch)           # pitch angle in radians.
    r = ranges * 1000.0                 # distances to gates in meters.

    x = r * (np.cos(rot + roll) * np.sin(drift) * np.cos(tilt) *
             np.sin(pitch) + np.cos(drift) * np.sin(rot + roll) *
             np.cos(tilt) - np.sin(drift) * np.cos(pitch) * np.sin(tilt))
    y = r * (-1. * np.cos(rot + roll) * np.cos(drift) * np.cos(tilt) *
             np.sin(pitch) + np.sin(drift) * np.sin(rot + roll) *
             np.cos(tilt) + np.cos(drift) * np.cos(pitch) * np.sin(tilt))
    z = (r * np.cos(pitch) * np.cos(tilt) * np.cos(rot + roll) +
         np.sin(pitch) * np.sin(tilt))
    return x, y, z


def antenna_to_cartesian_earth_relative(
        ranges, rot, roll, heading, tilt, pitch):
    """
    Calculate earth-relative Cartesian coordinates from radar coordinates

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    roll : array
        Roll angle of the radar in degrees.
    heading : array
        Heading (compass) angle of the radar in degrees clockwise from north.
    tilt : array
        Tilt angle of the radar in degrees.
    pitch : array
        Pitch angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    earth-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    roll = np.radians(roll)             # roll angle in radians.
    heading = np.radians(heading)       # drift angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    pitch = np.radians(pitch)           # pitch angle in radians.
    r = ranges * 1000.0                 # distances to gates in meters.

    x = r * (-1. * np.cos(rot + roll) * np.sin(heading) * np.cos(tilt) *
             np.sin(pitch) + np.cos(heading) * np.sin(rot + roll) *
             np.cos(tilt) + np.sin(heading) * np.cos(pitch) * np.sin(tilt))
    y = r * (-1. * np.cos(rot + roll) * np.cos(heading) * np.cos(tilt) *
             np.sin(pitch) - np.sin(heading) * np.sin(rot + roll) *
             np.cos(tilt) + np.cos(heading) * np.cos(pitch) * np.sin(tilt))
    z = (r * np.cos(pitch) * np.cos(tilt) * np.cos(rot + roll) +
         np.sin(pitch) * np.sin(tilt))
    return x, y, z


def antenna_to_cartesian_aircraft_relative(ranges, rot, tilt):
    """
    Calculate aircraft-relative Cartesian coordinates from radar coordinates.

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    tilt : array
        Tilt angle of the radar in degrees.

    Returns
    -------
    X, Y, Z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    earth-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    r = ranges * 1000.0                 # distances to gates in meters.
    x = r * np.cos(tilt) * np.sin(rot)
    y = r * np.sin(tilt)
    z = r * np.cos(rot) * np.cos(tilt)
    return x, y, z


def geographic_to_cartesian(lon, lat, projparams):
    """
    Geographic to Cartesian coordinate transform.

    Transform a set of Geographic coordinate (lat, lon) to a
    Cartesian/Cartographic coordinate (x, y) using pyproj or a build in
    Azimuthal equidistant projection.

    Parameters
    ----------
    lon, lat : array-like
        Geographic coordinates in degrees.
    projparams : dict or str
        Projection parameters passed to pyproj.Proj. If this parameter is a
        dictionary with a 'proj' key equal to 'pyart_aeqd' then a azimuthal
        equidistant projection will be used that is native to Py-ART and
        does not require pyproj to be installed. In this case a non-default
        value of R can be specified by setting the 'R' key to the desired
        value.

    Returns
    -------
    x, y : array-like
        Cartesian coordinates in meters unless projparams defines a value for R
        in different units.

    """
    if isinstance(projparams, dict) and projparams.get('proj') == 'pyart_aeqd':
        # Use Py-ART's Azimuthal equidistance projection
        lon_0 = projparams['lon_0']
        lat_0 = projparams['lat_0']
        if 'R' in projparams:
            R = projparams['R']
            x, y = geographic_to_cartesian_aeqd(lon, lat, lon_0, lat_0, R)
        else:
            x, y = geographic_to_cartesian_aeqd(lon, lat, lon_0, lat_0)
    else:
        # Use pyproj for the projection
        # check that pyproj is available
        if not _PYPROJ_AVAILABLE:
            raise MissingOptionalDependency(
                "PyProj is required to use geographic_to_cartesian "
                "with a projection other than pyart_aeqd but it is not "
                "installed")
        proj = pyproj.Proj(projparams)
        x, y = proj(lon, lat, inverse=False)
    return x, y


def geographic_to_cartesian_aeqd(lon, lat, lon_0, lat_0, R=6370997.):
    """
    Azimuthal equidistant geographic to Cartesian coordinate transform.

    Transform a set of geographic coordinates (lat, lon) to
    Cartesian/Cartographic coordinates (x, y) using a azimuthal equidistant
    map projection [1].

    .. math::

        x = R * k * \\cos(lat) * \\sin(lon - lon_0)

        y = R * k * [\\cos(lat_0) * \\sin(lat) -
                     \\sin(lat_0) * \\cos(lat) * \\cos(lon - lon_0)]

        k = c / \\sin(c)

        c = \\arccos(\\sin(lat_0) * \\sin(lat) +
                     \\cos(lat_0) * \\cos(lat) * \\cos(lon - lon_0))

    Where x, y are the Cartesian position from the center of projection;
    lat, lon the corresponding latitude and longitude; lat_0, lon_0 are the
    latitude and longitude of the center of the projection; R is the radius of
    the earth (defaults to ~6371 km).

    Parameters
    ----------
    lon, lat : array-like
        Longitude and latitude coordinates in degrees.
    lon_0, lat_0 : float
        Longitude and latitude, in degrees, of the center of the projection.
    R : float, optional
        Earth radius in the same units as x and y. The default value is in
        units of meters.

    Returns
    -------
    x, y : array
        Cartesian coordinates in the same units as R, typically meters.

    References
    ----------
    .. [1] Snyder, J. P. Map Projections--A Working Manual. U. S. Geological
        Survey Professional Paper 1395, 1987, pp. 191-202.

    """
    lon = np.atleast_1d(np.asarray(lon))
    lat = np.atleast_1d(np.asarray(lat))

    lon_rad = np.deg2rad(lon)
    lat_rad = np.deg2rad(lat)

    lat_0_rad = np.deg2rad(lat_0)
    lon_0_rad = np.deg2rad(lon_0)

    lon_diff_rad = lon_rad - lon_0_rad

    # calculate the arccos after ensuring all values in valid domain, [-1, 1]
    arg_arccos = (np.sin(lat_0_rad) * np.sin(lat_rad) +
                  np.cos(lat_0_rad) * np.cos(lat_rad) * np.cos(lon_diff_rad))
    arg_arccos[arg_arccos > 1] = 1
    arg_arccos[arg_arccos < -1] = -1
    c = np.arccos(arg_arccos)
    with warnings.catch_warnings():
        # division by zero may occur here but is properly addressed below so
        # the warnings can be ignored
        warnings.simplefilter("ignore", RuntimeWarning)
        k = c / np.sin(c)
    # fix cases where k is undefined (c is zero), k should be 1
    k[c == 0] = 1

    x = R * k * np.cos(lat_rad) * np.sin(lon_diff_rad)
    y = R * k * (np.cos(lat_0_rad) * np.sin(lat_rad) -
                 np.sin(lat_0_rad) * np.cos(lat_rad) * np.cos(lon_diff_rad))
    return x, y


def cartesian_to_geographic(x, y, projparams):
    """
    Cartesian to Geographic coordinate transform.

    Transform a set of Cartesian/Cartographic coordinates (x, y) to a
    geographic coordinate system (lat, lon) using pyproj or a build in
    Azimuthal equidistant projection.

    Parameters
    ----------
    x, y : array-like
        Cartesian coordinates in meters unless R is defined in different units
        in the projparams parameter.
    projparams : dict or str
        Projection parameters passed to pyproj.Proj. If this parameter is a
        dictionary with a 'proj' key equal to 'pyart_aeqd' then a azimuthal
        equidistant projection will be used that is native to Py-ART and
        does not require pyproj to be installed. In this case a non-default
        value of R can be specified by setting the 'R' key to the desired
        value.

    Returns
    -------
    lon, lat : array
        Longitude and latitude of the Cartesian coordinates in degrees.

    """
    if isinstance(projparams, dict) and projparams.get('proj') == 'pyart_aeqd':
        # Use Py-ART's Azimuthal equidistance projection
        lon_0 = projparams['lon_0']
        lat_0 = projparams['lat_0']
        if 'R' in projparams:
            R = projparams['R']
            lon, lat = cartesian_to_geographic_aeqd(x, y, lon_0, lat_0, R)
        else:
            lon, lat = cartesian_to_geographic_aeqd(x, y, lon_0, lat_0)
    else:
        # Use pyproj for the projection
        # check that pyproj is available
        if not _PYPROJ_AVAILABLE:
            raise MissingOptionalDependency(
                "PyProj is required to use cartesian_to_geographic "
                "with a projection other than pyart_aeqd but it is not "
                "installed")
        if pyproj.CRS('EPSG:21781').to_dict() == projparams:
            # For swiss coords use faster Swisstopo functions
            lon, lat, _ = swissCH1903_to_wgs84(x, y, 0)
        else:
            proj = pyproj.Proj(projparams)
            lon, lat = proj(x, y, inverse=True)
    return lon, lat


def cartesian_vectors_to_geographic(x, y, projparams, edges=False):
    """
    Cartesian vectors to Geographic coordinate transform.

    Transform a set of Cartesian/Cartographic coordinate vectors (x, y) to a
    geographic coordinate system (lat, lon) using pyproj or a build in
    Azimuthal equidistant projection finding the coordinates edges in
    Cartesian space if requested.

    Parameters
    ----------
    x, y : array 1D.
        Cartesian coordinate vectors in meters unless R is defined in
        different units in the projparams parameter.
    projparams : dict or str
        Projection parameters passed to pyproj.Proj. If this parameter is a
        dictionary with a 'proj' key equal to 'pyart_aeqd' then a azimuthal
        equidistant projection will be used that is native to Py-ART and
        does not require pyproj to be installed. In this case a
        non-default value of R can be specified by setting the 'R' key to the
        desired value.
    edges : bool, optional
        True to calculate the coordinates of the geographic edges by
        interpolating between Cartesian points and extrapolating at the
        boundaries. False to calculate the coordinate centers.

    Returns
    -------
    lon, lat : array
        Longitude and latitude of the Cartesian coordinates in degrees.

    """
    if edges:
        if len(x) > 1:
            x = _interpolate_axes_edges(x)
        if len(y) > 1:
            y = _interpolate_axes_edges(y)
    x, y = np.meshgrid(x, y)
    return cartesian_to_geographic(x, y, projparams)


def cartesian_to_geographic_aeqd(x, y, lon_0, lat_0, R=6370997.):
    """
    Azimuthal equidistant Cartesian to geographic coordinate transform.

    Transform a set of Cartesian/Cartographic coordinates (x, y) to
    geographic coordinate system (lat, lon) using a azimuthal equidistant
    map projection [1].

    .. math::

        lat = \\arcsin(\\cos(c) * \\sin(lat_0) +
                       (y * \\sin(c) * \\cos(lat_0) / \\rho))

        lon = lon_0 + \\arctan2(
            x * \\sin(c),
            \\rho * \\cos(lat_0) * \\cos(c) - y * \\sin(lat_0) * \\sin(c))

        \\rho = \\sqrt(x^2 + y^2)

        c = \\rho / R

    Where x, y are the Cartesian position from the center of projection;
    lat, lon the corresponding latitude and longitude; lat_0, lon_0 are the
    latitude and longitude of the center of the projection; R is the radius of
    the earth (defaults to ~6371 km). lon is adjusted to be between -180 and
    180.

    Parameters
    ----------
    x, y : array-like
        Cartesian coordinates in the same units as R, typically meters.
    lon_0, lat_0 : float
        Longitude and latitude, in degrees, of the center of the projection.
    R : float, optional
        Earth radius in the same units as x and y. The default value is in
        units of meters.

    Returns
    -------
    lon, lat : array
        Longitude and latitude of Cartesian coordinates in degrees.

    References
    ----------
    .. [1] Snyder, J. P. Map Projections--A Working Manual. U. S. Geological
        Survey Professional Paper 1395, 1987, pp. 191-202.

    """
    x = np.atleast_1d(np.asarray(x))
    y = np.atleast_1d(np.asarray(y))

    lat_0_rad = np.deg2rad(lat_0)
    lon_0_rad = np.deg2rad(lon_0)

    rho = np.sqrt(x*x + y*y)
    c = rho / R

    with warnings.catch_warnings():
        # division by zero may occur here but is properly addressed below so
        # the warnings can be ignored
        warnings.simplefilter("ignore", RuntimeWarning)
        lat_rad = np.arcsin(np.cos(c) * np.sin(lat_0_rad) +
                            y * np.sin(c) * np.cos(lat_0_rad) / rho)
    lat_deg = np.rad2deg(lat_rad)
    # fix cases where the distance from the center of the projection is zero
    lat_deg[rho == 0] = lat_0

    x1 = x * np.sin(c)
    x2 = rho*np.cos(lat_0_rad)*np.cos(c) - y*np.sin(lat_0_rad)*np.sin(c)
    lon_rad = lon_0_rad + np.arctan2(x1, x2)
    lon_deg = np.rad2deg(lon_rad)
    # Longitudes should be from -180 to 180 degrees
    lon_deg[lon_deg > 180] -= 360.
    lon_deg[lon_deg < -180] += 360.

    return lon_deg, lat_deg


def cartesian_to_antenna(x, y, z):
    """
    Returns antenna coordinates from Cartesian coordinates.

    Parameters
    ----------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Returns
    -------
    ranges : array
        Distances to the center of the radar gates (bins) in m.
    azimuths : array
        Azimuth angle of the radar in degrees. [-180., 180]
    elevations : array
        Elevation angle of the radar in degrees.

    """
    ranges = np.sqrt(x ** 2. + y ** 2. + z ** 2.)
    elevations = np.rad2deg(np.arctan(z / np.sqrt(x ** 2. + y ** 2.)))
    azimuths = np.rad2deg(np.arctan2(x, y))  # [-180, 180]
    azimuths[azimuths < 0.] += 360.  # [0, 360]

    return ranges, azimuths, elevations


def add_2d_latlon_axis(grid, **kwargs):
    """
    Add the latitude and longitude for grid points in the y, x plane.

    Adds a **latitude** and **longitude** dictionary to the axes attribute
    of a provided grid.  Addition is done in-place, nothing is returned from
    this function.  These dictionaries contain 2D arrays which specify the
    latitude and longitude of every point in the y, x plane.

    If available, the conversion is done using basemap.pyproj, extra arguments
    are passed to pyproj.Proj. If not available, an internal spherical
    azimuthal equidistant transformation is is used.

    Parameters
    ----------
    grid: grid object
        Cartesian grid object containing the 1d axes "x_disp", "y_disp" and
        scalar axes 'lat', 'lon'.
    kwargs: Pyproj options
        Options to be passed to Proj. If projection is not specified here it
        uses proj='aeqd' (azimuthal equidistant)

    Notes
    -----
    If Basemap is not available, calculation of the latitude, longitude is
    done using a azimuthal equidistant projection projection [1].
    It uses the mean radius of earth (6371 km)

    .. math::

        c = \\sqrt(x^2 + y^2)/R

        azi = \\arctan2(y,x) \\text{  # from east to north}

        lat = \\arcsin(\\cos(c)*\\sin(lat0)+\\sin(azi)*\\sin(c)*\\cos(lat0))

        lon = \\arctan2(\\cos(azi)*\\sin(c),\\cos(c)*\\cos(lat0)-
                        \\sin(azi)*\\sin(c)*\\sin(lat0)) + lon0

    Where x, y are the Cartesian position from the center of projection;
    lat, lon the corresponding latitude and longitude; lat0, lon0 the latitude
    and longitude of the center of the projection; R the mean radius of the
    earth (6371 km)

    References
    ----------
    .. [1] Snyder, J. P. Map Projections--A Working Manual. U. S. Geological
        Survey Professional Paper 1395, 1987, pp. 191-202.

    """
    warnings.warn(
        "add_2d_latlon_axis is deprecated and will be removed in a " +
        "future version of Py-ART", DeprecationWarning)
    try:
        from mpl_toolkits.basemap import pyproj
        if 'proj' not in kwargs:
            kwargs['proj'] = 'aeqd'
        x, y = np.meshgrid(
            grid.axes["x_disp"]['data'], grid.axes["y_disp"]['data'])
        b = pyproj.Proj(lat_0=grid.axes["lat"]['data'][0],
                        lon_0=grid.axes["lon"]['data'][0], **kwargs)
        lon, lat = b(x, y, inverse=True)
    except ImportError:
        warnings.warn('No basemap found, using internal implementation '
                      'for converting azimuthal equidistant to latlon')
        # azimutal equidistant projetion to latlon
        R = 6371.0 * 1000.0     # radius of earth in meters.

        x, y = np.meshgrid(grid.axes["x_disp"]['data'],
                           grid.axes["y_disp"]['data'])

        c = np.sqrt(x*x + y*y) / R
        phi_0 = grid.axes["lat"]['data'] * PI / 180
        azi = np.arctan2(y, x)  # from east to north

        lat = np.arcsin(np.cos(c) * np.sin(phi_0) +
                        np.sin(azi) * np.sin(c) * np.cos(phi_0)) * 180 / PI
        lon = (np.arctan2(np.cos(azi) * np.sin(c), np.cos(c) * np.cos(phi_0) -
                          np.sin(azi) * np.sin(c) * np.sin(phi_0)) * 180 /
               PI + grid.axes["lon"]['data'])
        lon = np.fmod(lon + 180, 360) - 180

    lat_axis = {
        'data':  lat,
        'long_name': 'Latitude for points in Cartesian system',
        'axis': 'YX',
        'units': 'degree_N',
        'standard_name': 'latitude',
    }

    lon_axis = {
        'data': lon,
        'long_name': 'Longitude for points in Cartesian system',
        'axis': 'YX',
        'units': 'degree_E',
        'standard_name': 'longitude',
    }

    grid.axes["latitude"] = lat_axis
    grid.axes["longitude"] = lon_axis


def corner_to_point(corner, point):
    """
    Return the x, y distances in meters from a corner to a point.

    Assumes a spherical earth model.

    This function is Deprecated, use the :py:func:`geographic_to_cartesian`
    function as a replacement.

    Parameters
    ----------
    corner : (float, float)
        Latitude and longitude in degrees of the corner.
    point : (float, float)
        Latitude and longitude in degrees of the point.

    Returns
    -------
    x, y : floats
        Distances from the corner to the point in meters.

    """
    warnings.warn(
        "corner_to_point is deprecated and will be removed in a future " +
        "version of Py-ART.\n" +
        "Additionally use of this function is discourage, the " +
        "geographic_to_cartesian function produces similar results while " +
        "allowing the map projection to be specified. ", DeprecationWarning)
    Re = 6371.0 * 1000.0
    Rc = _ax_radius(point[0], units='degrees')
    # print Rc/Re
    y = ((point[0] - corner[0]) / 360.0) * PI * 2.0 * Re
    x = ((point[1] - corner[1]) / 360.0) * PI * 2.0 * Rc
    return x, y


def swissCH1903_to_wgs84(chy, chx, chh, no_altitude_transform=False):
    """
    Convert swiss coordinates (CH1903 / LV03) to WGS84 coordinates

    The formulas for the coordinates transformation are taken from:
    "Formeln und Konstanten für die Berechnung der Schweizerischen
    schiefachsigen Zylinderprojektion und der Transformation
    zwischen Koordinatensystemen", chapter 4. "Näherungslösungen
    CH1903 <=> WGS84"
    Bundesamt für Landestopografie swisstopo (http://www.swisstopo.admin.ch),
    Oktober 2008

    Test example
    ------------
    wgs84 input:
        latitude  : 46 deg 2' 38.87''
        longitude : 8 deg 43' 49.79''
        altitude  : 650.60 m
    Result swiss CH1903:
        chy = 699 999.76  (700000)
        chx =  99 999.97  (100000)
        chh = 600.05      (600)

    Parameters
    ----------
    chy, chx : array-like
        Geographic coordinates CH1903 in meters. (chy = W-E, chx = S-N)
    chh : array-like
        Altitude in m
    no_altitude_transform : bool
        If set, do not convert altitude

    Returns
    -------
    lon, lat, chh : array-like
       Longitude and Latitude in WGS84 coordinates
    alt : array-like
        Altitude in meters

    """
    
    # 1. Axiliary values (% Bern)
    y_aux = (chy - 600000) / 1000000
    x_aux = (chx - 200000) / 1000000
    lat = (16.9023892 + (3.238272 * x_aux)) + \
            - (0.270978 * y_aux ** 2) + \
            - (0.002528 * x_aux ** 2) + \
            - (0.0447 * y_aux ** 2 * x_aux) + \
            - (0.0140 * x_aux ** 3)
                                    
    # Unit 10000" to 1" and convert seconds to degrees (dec)
    lat = (lat * 100) / 36
    
    lng = (2.6779094 + (4.728982 * y_aux)) + \
                + (0.791484 * y_aux * x_aux) + \
                + (0.1306 * y_aux * x_aux ** 2) + \
                - (0.0436 * y_aux ** 3)
    # Unit 10000" to 1" and convert seconds to degrees (dec)
    lng = (lng * 100) / 36

    if no_altitude_transform:
        h = chh
    else:
        h = (chh + 49.55) - (12.60 * y_aux) - (22.64 * x_aux)
        
    return (lng, lat, h)

def wgs84_to_swissCH1903(lon, lat, alt, no_altitude_transform=False):
    """
    Convert WGS84 coordinates to swiss coordinates (CH1903 / LV03)

    The formulas for the coordinates transformation are taken from:
    "Formeln und Konstanten für die Berechnung der Schweizerischen
    schiefachsigen Zylinderprojektion und der Transformation
    zwischen Koordinatensystemen", chapter 4. "Näherungslösungen
    CH1903 <=> WGS84"
    Bundesamt für Landestopografie swisstopo (http://www.swisstopo.admin.ch),
    Oktober 2008

    Test example
    ------------
    wgs84 input:
        latitude  : 46 deg 2' 38.87''
        longitude : 8 deg 43' 49.79''
        altitude  : 650.60 m
    Result swiss CH1903:
        chy = 699 999.76  (700000)
        chx =  99 999.97  (100000)
        chh = 600.05      (600)

    Parameters
    ----------
    lon, lat : array-like
        Geographic coordinates WGS84 in degrees.
    alt : array-like
        Altitude in m
    no_altitude_transform : bool
        If set, do not convert altitude

    Returns
    -------
    chy, chx, chh : array-like
       Coordinates in swiss CH1903 coordinates in meter

    """

    # 1. Transform longitude and latitude from [deg] to [angular seconds]
    phi_sec = lat * 3600.0
    lambda_sec = lon * 3600.0

    # 2. Calculate auxiliary quantities:
    phi = (phi_sec - 169028.66) / 10000.0
    lam = (lambda_sec - 26782.5) / 10000.0

    # 3. Use approximation formulas for chx, chy and altitude
    chy = \
        600072.37 + \
        211455.93 * lam - \
        10938.51 * lam * phi - \
        0.36 * lam * phi**2 - \
        44.54 * lam**3

    chx = \
        200147.07 + \
        308807.95 * phi + \
        3745.25 * lam**2 + \
        76.63 * phi**2 - \
        194.56 * lam**2 * phi + \
        119.79 * phi**3

    if (no_altitude_transform):
        chh = alt
    else:
        chh = alt - 49.55 + \
              2.73 * lam + \
              6.94 * phi

    return (chy, chx, chh)


def _ax_radius(lat, units='radians'):
    """
    Return the radius of a constant latitude circle for a given latitude.

    Parameters
    ----------
    lat : float
        Latitude at which to calculate constant latitude circle (parallel)
        radius.
    units : 'radians' or 'degrees'
        Units of lat, either 'radians' or 'degrees'.

    Returns
    -------
    R : float
        Radius in meters of a constant latitude circle (parallel).

    """
    Re = 6371.0 * 1000.0
    if units == 'degrees':
        const = PI / 180.0
    else:
        const = 1.0
    R = Re * np.sin(PI / 2.0 - abs(lat * const))
    return R



"""
pyrad.correct.sunlib
====================

Library to deal with sun measurements

.. autosummary::
    :toctree: generated/

    sun_position_pysolar
    sun_position_mfr
    equation_of_time
    hour_angle
    solar_declination
    refraction_correction
    gas_att_sun
    gauss_fit
    retrieval_result
    sun_power
    ptoa_to_sf
    solar_flux_lookup
    scanning_losses

"""

import datetime
from copy import deepcopy
from warnings import warn

from numpy import pi, sin, cos, arcsin, arccos, sqrt, floor
from numpy.linalg import LinAlgError
import numpy as np
from scipy.special import erf

try:
    import pysolar
    _PYSOLAR_AVAILABLE = True
except ImportError:
    _PYSOLAR_AVAILABLE = False

def sun_position_pysolar(dt, lat, lon, elevation=0.):
    """
    obtains the sun position in antenna coordinates using the pysolar
    library.

    Parameters
    ----------
    dt : datetime object
        the time when to look for the sun
    lat, lon : float
        latitude and longitude of the sensor in degrees

    Returns
    -------
    el, az : float
        elevation and azimuth angles of the sun respect to the sensor in
        degrees

    """
    # Make the date time zone aware
    dt_aux = deepcopy(dt)
    dt_aux = dt_aux.replace(tzinfo=datetime.timezone.utc)
    az, el = pysolar.solar.get_position(lat, lon, dt_aux, elevation=elevation)

    return (el, az)


def sun_position_mfr(dt, lat_deg, lon_deg, refraction=True):
    """
    Calculate the sun position for the given time (dt) at the given position
    (lat, lon).

    Parameters
    ----------
    dt : datetime object
        the time when to look for the sun
    lat_deg, lon_deg: floats
        latitude and longitude in degrees
    refraction : boolean
        whether to correct for refraction or not

    Returns
    -------
    elev_sun, azim_sun : floats
        elevation and azimuth angles of the sun respect to the sensor in
        degrees

    """
    lat = lat_deg*pi/180.
    lon = lon_deg*pi/180.

    secs_since_midnight = (
        (dt - dt.replace(
            hour=0, minute=0, second=0, microsecond=0)).total_seconds())
    htime = secs_since_midnight/3600.
    dayjul = (
        (dt - dt.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0)).days +
        1)

    eqt = equation_of_time(dayjul)  # [h]
    hang = hour_angle(htime, lon, eqt)  # [rad]
    sdec = solar_declination(dayjul, htime)  # [rad]

    elev_sun = (
        arcsin(sin(lat)*sin(sdec)+cos(lat)*cos(sdec)*cos(hang))*180./pi)
    azim_sun = (
        arccos((sin(lat)*cos(sdec)*cos(hang)-cos(lat)*sin(sdec)) /
               cos(elev_sun*pi/180.))*180./pi)

    if hang < 0:
        azim_sun = 180. - azim_sun  # morning
    else:
        azim_sun = 180. + azim_sun  # afternoon

    if refraction:
        elev_sun += refraction_correction(elev_sun)

    return (elev_sun, azim_sun)


def equation_of_time(dayjul):
    """
    Computes the solar hour for a given julian day.

    Parameters
    ----------
    dayjul : double
        julian date

    Returns
    -------
    eqt : float
        hour

    """
    temp_cos = [-0.00720, 0.0528, 0.0012]
    temp_sin = [0.12290, 0.1565, 0.0041]
    omega = 2.*pi/365.2425  # earth mean angular orbital velocity [rad/day]

    eqt = 0.
    for ii in range(3):
        z = dayjul*omega*(ii+1)
        eqt += (temp_cos[ii]*cos(z)+temp_sin[ii]*sin(z))

    return -eqt  # [h]


def hour_angle(htime, lon, eqt):
    """
    Computes the solar angle at a particular time.

    Parameters
    ----------
    htime : double
        time in seconds since midnight
    lon : float
        longitude in degrees
    eqt : float
        solar time

    Returns
    -------
    angle : float
        the solar angle in radiants

    """
    return (htime+12./pi*lon+eqt-12.)*pi/12.  # [rad]


def solar_declination(dayjul, htime):
    """
    Computes the solar declination.

    Parameters
    ----------
    dayjul : double
        julian date
    htime : double
        time in seconds since midnight

    Returns
    -------
    angle : float
        the solar declination in radiants

    """
    omega = 2.*pi/365.2425  # earth mean angular orbital velocity [rad/day]
    correction = [0., 3., 6., 9., 12., 13., 12., 10., 10., 8., 6., 3., 1.,
                  -2., -3., -4., -5., -5., -3., -3., -1., -1., 0., 0., 1., 3.,
                  6.]

    z = dayjul * omega
    x = 0.33281 - 22.984 * cos(z) - 0.3499 * cos(2.*z) - 0.1398 * cos(3. * z)
    y = 3.7872 * sin(z) + 0.03205 * sin(2. * z) + 0.07187 * sin(3. * z)

    fortnight = int(floor(dayjul/15)) + 1
    day_fortnight = dayjul-(fortnight-1)*15
    corr1 = (
        (correction[fortnight]+day_fortnight/15. *
         (correction[fortnight+1]-correction[fortnight]))/60.)
    delta1 = x + y + corr1

    z = (dayjul+1) * omega
    x = 0.33281 - 22.984 * cos(z) - 0.3499 * cos(2.*z) - 0.1398 * cos(3. * z)
    y = 3.7872 * sin(z) + 0.03205 * sin(2. * z) + 0.07187 * sin(3. * z)

    fortnight = int(floor(dayjul+1)/15) + 1
    day_fortnight = (dayjul+1)-(fortnight-1)*15
    corr2 = (
        (correction[fortnight]+day_fortnight/15. *
         (correction[fortnight+1]-correction[fortnight]))/60.)
    delta2 = x + y + corr2

    return (delta1+(delta2-delta1)*htime/24.)*pi/180.  # [rad]


def refraction_correction(es_deg):
    """
    Computes the correction that has to be applied to the sun elevation angle
    to account for refraction

    Parameters
    ----------
    es_deg : float
        sun elevation in degrees

    Returns
    -------
    refr : float
        the correction due to refraction in degrees

    References
    ----------
    Holleman & Huuskonen, 2013: analytical formulas for refraction of
    radiowaves from exoatmospheric sources, radio science, vol. 48, 226-231

    """
    if es_deg < -0.77:
        return 0.0
    es_rad = es_deg*pi/180.
    k = 5./4.  # effective earth radius factor (typically 4/3)
    n = 313.   # surface refractivity
    no = n*1e-6 + 1.
    refr = (
        ((k-1.)/(2.*k-1.)*cos(es_rad) *
         (sqrt((sin(es_rad))**2.+(4.*k-2.)/(k-1.)*(no-1.)) -
          sin(es_rad)))*180./pi)

    return refr


def gas_att_sun(es_deg, attg):
    """
    Computes the attenuation suffered by the sun signal through the atmosphere

    Parameters
    ----------
    es_deg : float
        sun elevation in degrees
    attg : float
        1-way gas attenuation in dB/km

    Returns
    -------
    gas_att_sun : float
        the sun attenuation in dB

    """
    r43 = 4./3.*6371  # effective earth radius [km]
    z0 = 8.4          # equivalent height of the atmosphere [km]
    return attg*(r43*sqrt((sin(es_deg*pi/180.))**2.+2.*z0/r43+(z0/r43)**2) -
                 r43*sin(es_deg*pi/180.))


def gauss_fit(az_data, az_ref, el_data, el_ref, sunhits, npar, degree=True,
              do_elcorr=True):
    """
    estimates a gaussian fit of sun hits data

    Parameters
    ----------
    az_data, el_data : float array
        azimuth and elevation radar data
    az_ref, el_ref : float array
        azimuth and elevation sun data
    sunhits : float array
        sun hits data
    npar : int
        number of parameters of the fit
    degree : boolean
        boolean indicating whether the data is in degree or radians
    do_elcorr : boolean
        indicates whether azimuth data is corrected so that azimuth
        differences are invalid with elevation

    Returns
    -------
    par : 1D float array
        the fit parameters
    alpha: 2D float array
        the matrix used in the fit
    beta: 1D float array
        the vector used in the fit

    """
    nhits = len(az_data)

    el_corr = 1.
    if do_elcorr:
        if degree:
            el_corr = np.ma.cos(el_data*np.pi/180.)
        else:
            el_corr = np.ma.cos(el_data)

    basis = np.ma.zeros((npar, nhits))
    basis[0, :] = 1.
    basis[1, :] = (az_data-az_ref)*el_corr
    basis[2, :] = el_data-el_ref

    if npar == 5:
        basis[3, :] = basis[1, :]*basis[1, :]
        basis[4, :] = basis[2, :]*basis[2, :]

    alpha = np.ma.zeros((npar, npar))
    beta = np.ma.zeros(npar)
    for hit in range(nhits):
        for ipar in range(npar):
            for jpar in range(npar):
                alpha[jpar, ipar] += basis[jpar, hit] * basis[ipar, hit]
            beta[ipar] += sunhits[hit] * basis[ipar, hit]

    try:
        alphainv = np.linalg.inv(alpha)
        par = np.ma.dot(alphainv, beta)

        return par, alpha, beta

    except LinAlgError:
        warn('Unable to perform Guassian fit of sun hits data')
        return None, None, None


def retrieval_result(sunhits, alpha, beta, par, npar):
    """
    computes the physical parameters of the sun retrieval from the results of
    a Gaussian fit.

    Parameters
    ----------
    sunhits : float array
        sun hits data
    alpha: 2D float array
        the matrix used in the fit
    beta: 1D float array
        the vector used in the fit
    par : 1D float array
        the fit parameters
    npar : int
        number of parameters of the fit

    Returns
    -------
    val, val_std : float
        retrieved value and its standard deviation
    az_bias, el_bias : float
        retrieved azimuth and elevation antenna bias respect to the sun
        position
    az_width, el_width : float
        retrieved azimuth and elevation antenna widths

    """
    nhits = len(sunhits)

    val = (
        par[0]-0.25*np.ma.power(par[1], 2.)/par[3] -
        0.25*np.ma.power(par[2], 2.)/par[4])

    az_bias = -0.5*par[1]/par[3]
    el_bias = -0.5*par[2]/par[4]

    coeff = -40.*np.ma.log10(2.)
    az_width = np.ma.sqrt(coeff/par[3])
    el_width = np.ma.sqrt(coeff/par[4])

    val_std = np.ma.sum(np.ma.power(sunhits, 2.))-2.*np.ma.sum(par*beta)
    for ipar in range(npar):
        for jpar in range(npar):
            val_std += par[ipar]*par[jpar]*alpha[ipar, jpar]
    val_std = np.ma.sqrt(val_std/(nhits-npar))

    return val, val_std, az_bias, el_bias, az_width, el_width


def sun_power(solar_flux, pulse_width, wavelen, antenna_gain, angle_step,
              beamwidth, coeff_band=1.2):
    """
    computes the theoretical sun power detected at the antenna [dBm] as it
    would be without atmospheric attenuation (sun power at top of the
    atmosphere) for a given solar flux and radar characteristics

    Parameters
    ----------
    solar_flux : float array
        the solar fluxes measured at 10.7 cm wavelength [10e-22 W/(m2 Hz)]
    pulse_width : float
        pulse width [s]
    wavelen : float
        radar wavelength [m]
    antenna_gain : float
        the antenna gain [dB]
    angle_step : float
        integration angle [deg]
    beamwidth : float
        3 dB-beamwidth [deg]
    coeff_band : float
        multiplicative coefficient applied to the inverse of the pulse width
        to get the effective bandwidth

    Returns
    -------
    pwr_det : float array
        the detected power

    References
    ----------
    Altube P., J. Bech, O. Argemi, T. Rigo, 2015: Quality Control of Antenna
    Alignment and Receiver Calibration Using the Sun: Adaptation to Midrange
    Weather Radar Observations at Low Elevation Angles

    """
    g = np.power(10., 0.1*antenna_gain)
    b = coeff_band*1./pulse_width  # receiver bandwidth [Hz]

    aeff = g*wavelen**2./(4.*np.pi)  # effective area of the antenna [m2]

    # solar flux at given wavelength
    s0 = solar_flux_lookup(solar_flux, wavelen)

    ptoa = 10.*np.log10(0.5*b*aeff*s0*1e-19)  # sun power at TOA [dBm]

    # losses due to antenna beam width and scanning
    la = scanning_losses(angle_step, beamwidth)

    return ptoa-la


def ptoa_to_sf(ptoa, pulse_width, wavelen, antenna_gain, coeff_band=1.2):
    """
    Converts the sun power at the top of the atmosphere (in dBm) into solar
    flux.

    Parameters
    ----------
    ptoa : float
        sun power at the top of the amosphere. It already takes into account
        the correction for antenna polarization
    pulse_width : float
        pulse width [s]
    wavelen : float
        radar wavelength [m]
    antenna_gain : float
        the antenna gain [dB]
    coeff_band : float
        multiplicative coefficient applied to the inverse of the pulse width
        to get the effective bandwidth

    Returns
    -------
    s0 : float
        solar flux [10e-22 W/(m2 Hz)]

    References
    ----------
    Altube P., J. Bech, O. Argemi, T. Rigo, 2015: Quality Control of Antenna
    Alignment and Receiver Calibration Using the Sun: Adaptation to Midrange
    Weather Radar Observations at Low Elevation Angles

    """
    g = np.power(10., 0.1*antenna_gain)
    b = coeff_band*1./pulse_width  # receiver bandwidth [Hz]

    aeff = g*wavelen**2./(4.*np.pi)  # effective area of the antenna [m2]

    # solar flux in [10e-22 W/(m2 Hz)]
    s0 = np.power(10., 0.1*ptoa)*1e19/(b*aeff)

    return s0


def solar_flux_lookup(solar_flux, wavelen):
    """
    Given the observed solar flux at 10.7 cm wavelength, returns the solar
    flux at the given radar wavelength

    Parameters
    ----------
    solar_flux : float array
        the solar fluxes measured at 10.7 cm wavelength [10e-22 W/(m2 Hz)]
    wavelen : float
        radar wavelength [m]

    Returns
    -------
    s0 : float
        the radar flux at the radar wavelength [10e-22 W/(m2 Hz)]

    References
    ----------
    Altube P., J. Bech, O. Argemi, T. Rigo, 2015: Quality Control of Antenna
    Alignment and Receiver Calibration Using the Sun: Adaptation to Midrange
    Weather Radar Observations at Low Elevation Angles

    """
    # minimum flux
    mfu = [1980., 495., 255., 170., 126., 102., 88., 76., 72., 68., 64., 61.,
           58., 55., 54., 53., 52., 51., 50., 49., 48., 48., 47., 47., 47.,
           46., 46., 45., 45., 45.]

    # scale factor
    sfa = [0.67, 0.68, 0.69, 0.70, 0.71, 0.73, 0.78, 0.84, 0.96, 1.00, 1.00,
           0.98, 0.94, 0.90, 0.85, 0.80, 0.78, 0.77, 0.76, 0.75, 0.74, 0.73,
           0.72, 0.71, 0.70, 0.69, 0.68, 0.67, 0.66, 0.65]

    ind_w = int(wavelen*100.)-1  # table index
    s0 = sfa[ind_w]*(solar_flux-64.)+mfu[ind_w]  # solar flux at wavelen

    return s0


def scanning_losses(angle_step, beamwidth):
    """
    Given the antenna beam width and the integration angle, compute the
    losses due to the fact that the sun is not a point target and the antenna
    is scanning


    Parameters
    ----------
    angle_step : float
        integration angle [deg]
    beamwidth : float
        3 dB-beamwidth [deg]

    Returns
    -------
    la : float
        The losses due to the scanning of the antenna [dB positive]

    References
    ----------
    Altube P., J. Bech, O. Argemi, T. Rigo, 2015: Quality Control of Antenna
    Alignment and Receiver Calibration Using the Sun: Adaptation to Midrange
    Weather Radar Observations at Low Elevation Angles

    """
    delta_s = 0.57  # apparent diameter of radio sun [deg]

    # sun convoluted antenna beamwidth look up table according to
    # Altube et al. (2015) Table 2
    delta_b = np.asarray(
        [0.70, 0.75, 0.80, 0.85, 0.90, 0.95, 1.00, 1.10, 1.20, 1.30, 1.40,
         1.50])
    delta_c0 = np.asarray(
        [0.78, 0.83, 0.87, 0.92, 0.96, 1.01, 1.06, 1.15, 1.25, 1.34, 1.44,
         1.54])

    if beamwidth < delta_b[0] or beamwidth > delta_b[-1]:
        warn('Antenna beam width outside of range of valid antenna values ' +
             'used to calculate sun convoluted beamwidth. The nominal ' +
             'antenna beamwidth will be used instead.')
        delta_c = beamwidth
    else:
        ind_c = np.where(delta_b <= beamwidth)[0][-1]
        delta_c = (
            delta_c0[ind_c]+(beamwidth-delta_b[ind_c]) *
            (delta_c0[ind_c+1]-delta_c0[ind_c]) /
            (delta_b[ind_c+1]-delta_b[ind_c]))

    # losses due to scanning and antenna beamwidth
    l0 = 1./np.log(2.)*beamwidth**2./delta_s**2.*(
        1.-np.exp(-np.log(2.)*delta_s**2./beamwidth**2))
    la = -10.*np.log10(
        l0*np.sqrt(np.pi/(4.*np.log(2.)))*delta_c/angle_step *
        erf(np.sqrt(np.log(2.))*angle_step/delta_c))

    return la



"""
pyart.testing.sample_files
==========================

Sample radar files in a number of formats. Many of these files
are incomplete, they should only be used for testing, not production.

.. autosummary::
    :toctree: generated/

    MDV_PPI_FILE
    MDV_RHI_FILE
    CFRADIAL_PPI_FILE
    CFRADIAL_RHI_FILE
    CHL_RHI_FILE
    SIGMET_PPI_FILE
    SIGMET_RHI_FILE
    NEXRAD_ARCHIVE_MSG31_FILE
    NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE
    NEXRAD_ARCHIVE_MSG1_FILE
    NEXRAD_LEVEL3_MSG19
    NEXRAD_LEVEL3_MSG163
    NEXRAD_CDM_FILE
    UF_FILE
    INTERP_SOUNDE_FILE

"""

import os

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')

MDV_PPI_FILE = os.path.join(DATA_PATH, 'example_mdv_ppi.mdv')
MDV_RHI_FILE = os.path.join(DATA_PATH, 'example_mdv_rhi.mdv')
MDV_GRID_FILE = os.path.join(DATA_PATH, 'example_mdv_grid.mdv')
CFRADIAL_PPI_FILE = os.path.join(DATA_PATH, 'example_cfradial_ppi.nc')
CFRADIAL_RHI_FILE = os.path.join(DATA_PATH, 'example_cfradial_rhi.nc')
CHL_RHI_FILE = os.path.join(DATA_PATH, 'example_chl_rhi.chl')
SIGMET_PPI_FILE = os.path.join(DATA_PATH, 'example_sigmet_ppi.sigmet')
SIGMET_RHI_FILE = os.path.join(DATA_PATH, 'example_sigmet_rhi.sigmet')
NEXRAD_ARCHIVE_MSG31_FILE = os.path.join(
    DATA_PATH, 'example_nexrad_archive_msg31.bz2')
NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE = os.path.join(
    DATA_PATH, 'example_nexrad_archive_msg31_compressed.ar2v')
NEXRAD_ARCHIVE_MSG1_FILE = os.path.join(
    DATA_PATH, 'example_nexrad_archive_msg1.bz2')
# NEXRAD Level 3 file downloaded from NCDC with filenames:
# KBMX_SDUS54_N0RBMX_201501020205
# KBMX_SDUS84_N0KBMX_201501020205
NEXRAD_LEVEL3_MSG19 = os.path.join(DATA_PATH, 'example_nexrad_level3_msg19')
NEXRAD_LEVEL3_MSG163 = os.path.join(DATA_PATH, 'example_nexrad_level3_msg163')
NEXRAD_CDM_FILE = os.path.join(DATA_PATH, 'example_nexrad_cdm.bz2')
RHI_ML_FILE = os.path.join(DATA_PATH, 'MXPol-polar-20120929-064418-RHI-166_6.nc')
UF_FILE = os.path.join(DATA_PATH, 'example_uf_ppi.uf')
INTERP_SOUNDE_FILE = os.path.join(DATA_PATH, 'example_interpolatedsonde.cdf')
SONDE_FILE = os.path.join(DATA_PATH, 'example_arm_sonde.cdf')
_EXAMPLE_RAYS_FILE = os.path.join(DATA_PATH, 'example_rays.npz')



from __future__ import division, absolute_import, print_function

import sys
from nose import SkipTest

def test_import():
    if sys.version_info[0] >= 3:
        raise SkipTest("plot_directive not ported to Python 3 (use the one from Matplotlib instead)")
    import numpydoc.plot_directive

# No tests at the moment...



"""
pyart.exceptions
================

Custom Py-ART exceptions.

.. autosummary::
    :toctree: generated/

    MissingOptionalDependency
    DeprecatedAttribute
    DeprecatedFunctionName
    _deprecated_alias

"""

import warnings


class MissingOptionalDependency(Exception):
    """ Exception raised when a optional dependency is needed by not found. """
    pass


class DeprecatedAttribute(DeprecationWarning):
    """ Warning category for an attribute which has been renamed/moved.  """
    pass


class DeprecatedFunctionName(DeprecationWarning):
    """ Warning category for a function which has been renamed/moved. """
    pass


def _deprecated_alias(func, old_name, new_name):
    """

    A function for creating an alias to a renamed or moved function.

    Parameters
    ----------
    func : func
        The function which has been renamed or moved.
    old_name, new_name : str
        Name of the function before and after it was moved or renamed
        (with namespace if changed).

    Returns
    -------
    wrapper : func
        A wrapper version of func, which issues a DeprecatedFunctionName
        warning when the called.

    """
    def wrapper(*args, **kwargs):
        warnings.warn(
            ("{0} has been deprecated and will be removed in future " +
             "versions of Py-ART, pleases use {1}. ").format(
                 old_name, new_name), category=DeprecatedFunctionName)
        return func(*args, **kwargs)
    return wrapper



""" Unit Tests for Py-ART's correct/attenuation.py module. """

# python test_phase_attenuation.py
# to recreate the reference_rays.npz file.

import os

import pyart
import numpy as np
from numpy.testing import assert_allclose

PATH = os.path.dirname(__file__)
REFERENCE_RAYS_FILE = os.path.join(PATH, 'attenuation_rays.npz')
REFERENCE_RAYS_FILE_ZPHI = os.path.join(PATH, 'attenuation_rays_zphi.npz')
REFERENCE_RAYS_FILE_PHILINEAR = os.path.join(
    PATH, 'attenuation_rays_philinear.npz')


def test_attenuation():
    spec_at, cor_z = perform_attenuation()
    ref = np.load(REFERENCE_RAYS_FILE)
    assert_allclose(ref['spec_at'], spec_at['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['cor_z'], cor_z['data'].data, rtol=1e-2, atol=1e-3)


def perform_attenuation_zphi():
    """ Perform attenuation correction on a single ray radar. """
    radar = pyart.testing.make_single_ray_radar()
    a = radar.fields['reflectivity']['data']
    radar.fields['reflectivity']['data'] = np.ma.array(a)
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr = (
        pyart.correct.calculate_attenuation_zphi(radar, a_coef=0.06, beta=0.8,
                                                 fzl=4000.0, c=0.15917,
                                                 d=1.0804, doc=0.0,
                                                 temp_ref='fixed_fzl'))
    return spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr


def perform_attenuation_philinear():
    """ Perform attenuation correction on a single ray radar. """
    radar = pyart.testing.make_single_ray_radar()
    a = radar.fields['reflectivity']['data']
    radar.fields['reflectivity']['data'] = np.ma.array(a)
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr = (
        pyart.correct.calculate_attenuation_philinear(radar, pia_coef=0.06,
                                                      pida_coef=0.8,
                                                      fzl=4000.0, doc=0.0,
                                                      temp_ref='fixed_fzl'))
    return spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr


def perform_attenuation():
    """ Perform attenuation correction on a single ray radar. """
    radar = pyart.testing.make_single_ray_radar()
    a = radar.fields['reflectivity']['data']
    radar.fields['reflectivity']['data'] = np.ma.array(a)
    spec_at, cor_z = pyart.correct.calculate_attenuation(radar, 0.0)
    return spec_at, cor_z


def save_reference_rays(spec_at, cor_z):
    """ Save the phase processed rays to REFERENCE_RAY_FILE. """
    np.savez(REFERENCE_RAYS_FILE,
             spec_at=spec_at['data'], cor_z=cor_z['data'].data)


def save_reference_rays_zphi(spec_at, pia_dict, cor_z, spec_diff_at,
                             pida_dict, cor_zdr):
    """ Save the phase processed rays to REFERENCE_RAY_FILE_ZPHI. """
    np.savez(REFERENCE_RAYS_FILE_ZPHI,
             spec_at=spec_at['data'], cor_z=cor_z['data'].data,
             pia_dict=pia_dict['data'], spec_diff_at=spec_diff_at['data'],
             pida_dict=pida_dict['data'], cor_zdr=cor_zdr['data'])


def save_reference_rays_philinear(spec_at, pia_dict, cor_z, spec_diff_at,
                                  pida_dict, cor_zdr):
    """ Save the phase processed rays to REFERENCE_RAY_FILE_PHILINEAR. """
    np.savez(REFERENCE_RAYS_FILE_PHILINEAR,
             spec_at=spec_at['data'], cor_z=cor_z['data'].data,
             pia_dict=pia_dict['data'], spec_diff_at=spec_diff_at['data'],
             pida_dict=pida_dict['data'], cor_zdr=cor_zdr['data'])


if __name__ == "__main__":
    spec_at, cor_z = perform_attenuation()
    save_reference_rays(spec_at, cor_z)
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr = (
        perform_attenuation_zphi())
    save_reference_rays_zphi(spec_at, pia_dict, cor_z, spec_diff_at,
                             pida_dict, cor_zdr)
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr = (
        perform_attenuation_philinear())
    save_reference_rays_philinear(spec_at, pia_dict, cor_z, spec_diff_at,
                                  pida_dict, cor_zdr)
    print("Reference rays saved")


def test_specific_diff_attenuation():
    """ Perform attenuation correction on a single ray radar. """
    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr = (
        perform_attenuation_zphi())
    ref = np.load(REFERENCE_RAYS_FILE_ZPHI)
    assert_allclose(ref['spec_at'], spec_at['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['cor_z'], cor_z['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['pia_dict'], pia_dict['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['spec_diff_at'], spec_diff_at['data'], 
                    rtol=1e-2, atol=1e-3)
    assert_allclose(ref['pida_dict'], pida_dict['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['cor_zdr'], cor_zdr['data'], rtol=1e-2, atol=1e-3)

    spec_at, pia_dict, cor_z, spec_diff_at, pida_dict, cor_zdr = (
        perform_attenuation_philinear())
    ref = np.load(REFERENCE_RAYS_FILE_PHILINEAR)
    assert_allclose(ref['spec_at'], spec_at['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['cor_z'], cor_z['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['pia_dict'], pia_dict['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['spec_diff_at'], spec_diff_at['data'], 
                    rtol=1e-2, atol=1e-3)
    assert_allclose(ref['pida_dict'], pida_dict['data'], rtol=1e-2, atol=1e-3)
    assert_allclose(ref['cor_zdr'], cor_zdr['data'], rtol=1e-2, atol=1e-3)
